# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""SqlAlchemy implementation of `aiida.orm.implementation.backends.Backend`."""
from contextlib import contextmanager

from aiida.backends.sqlalchemy.models import base
from aiida.backends.sqlalchemy.queries import SqlaQueryManager
from aiida.backends.sqlalchemy.manager import SqlaBackendManager

from ..sql.backends import SqlBackend
from . import authinfos
from . import comments
from . import computers
from . import convert
from . import groups
from . import logs
from . import nodes
from . import querybuilder
from . import users

__all__ = ('SqlaBackend',)


class SqlaBackend(SqlBackend[base.Base]):
    """SqlAlchemy implementation of `aiida.orm.implementation.backends.Backend`."""

    def __init__(self):
        """Construct the backend instance by initializing all the collections."""
        self._authinfos = authinfos.SqlaAuthInfoCollection(self)
        self._comments = comments.SqlaCommentCollection(self)
        self._computers = computers.SqlaComputerCollection(self)
        self._groups = groups.SqlaGroupCollection(self)
        self._logs = logs.SqlaLogCollection(self)
        self._nodes = nodes.SqlaNodeCollection(self)
        self._query_manager = SqlaQueryManager(self)
        self._schema_manager = SqlaBackendManager()
        self._users = users.SqlaUserCollection(self)

    def migrate(self):
        self._schema_manager.migrate()

    @property
    def authinfos(self):
        return self._authinfos

    @property
    def comments(self):
        return self._comments

    @property
    def computers(self):
        return self._computers

    @property
    def groups(self):
        return self._groups

    @property
    def logs(self):
        return self._logs

    @property
    def nodes(self):
        return self._nodes

    @property
    def query_manager(self):
        return self._query_manager

    def query(self):
        return querybuilder.SqlaQueryBuilder(self)

    @property
    def users(self):
        return self._users

    @contextmanager
    def transaction(self):
        """Open a transaction to be used as a context manager.

        If there is an exception within the context then the changes will be rolled back and the state will be as before
        entering. Transactions can be nested.
        """
        session = self.get_session()
        nested = session.transaction.nested
        try:
            session.begin_nested()
            yield session
            session.commit()
        except Exception:
            session.rollback()
            raise
        finally:
            if not nested:
                # Make sure to commit the outermost session
                session.commit()

    @staticmethod
    def get_session():
        """Return a database session that can be used by the `QueryBuilder` to perform its query.

        :return: an instance of :class:`sqlalchemy.orm.session.Session`
        """
        from aiida.backends.sqlalchemy import get_scoped_session
        return get_scoped_session()

    # Below are abstract methods inherited from `aiida.orm.implementation.sql.backends.SqlBackend`

    def get_backend_entity(self, model):
        """Return a `BackendEntity` instance from a `DbModel` instance."""
        return convert.get_backend_entity(model, self)

    @contextmanager
    def cursor(self):
        """Return a psycopg cursor to be used in a context manager.

        :return: a psycopg cursor
        :rtype: :class:`psycopg2.extensions.cursor`
        """
        from aiida.backends import sqlalchemy as sa
        try:
            connection = sa.ENGINE.raw_connection()
            yield connection.cursor()
        finally:
            self.get_connection().close()

    def execute_raw(self, query):
        """Execute a raw SQL statement and return the result.

        :param query: a string containing a raw SQL statement
        :return: the result of the query
        """
        from sqlalchemy.exc import ResourceClosedError  # pylint: disable=import-error,no-name-in-module

        with self.transaction() as session:
            queryset = session.execute(query)

            try:
                results = queryset.fetchall()
            except ResourceClosedError:
                return None

        return results

    @staticmethod
    def get_connection():
        """Get the SQLA database connection

        :return: the SQLA database connection
        """
        from aiida.backends import sqlalchemy as sa
        return sa.ENGINE.raw_connection()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
""" Utility functions for export of AiiDA entities """
# pylint: disable=too-many-locals,too-many-branches,too-many-nested-blocks
import warnings

from aiida.orm import QueryBuilder, ProcessNode
from aiida.common.log import AIIDA_LOGGER, LOG_LEVEL_REPORT
from aiida.common.warnings import AiidaDeprecationWarning

from aiida.tools.importexport.common import exceptions
from aiida.tools.importexport.common.config import (
    file_fields_to_model_fields, entity_names_to_entities, get_all_fields_info
)

EXPORT_LOGGER = AIIDA_LOGGER.getChild('export')


def fill_in_query(partial_query, originating_entity_str, current_entity_str, tag_suffixes=None, entity_separator='_'):
    """
    This function recursively constructs QueryBuilder queries that are needed
    for the SQLA export function. To manage to construct such queries, the
    relationship dictionary is consulted (which shows how to reference
    different AiiDA entities in QueryBuilder.
    To find the dependencies of the relationships of the exported data, the
    get_all_fields_info (which described the exported schema and its
    dependencies) is consulted.
    """
    if not tag_suffixes:
        tag_suffixes = []

    relationship_dic = {
        'Node': {
            'Computer': 'with_computer',
            'Group': 'with_group',
            'User': 'with_user',
            'Log': 'with_log',
            'Comment': 'with_comment'
        },
        'Group': {
            'Node': 'with_node'
        },
        'Computer': {
            'Node': 'with_node'
        },
        'User': {
            'Node': 'with_node',
            'Group': 'with_group',
            'Comment': 'with_comment',
        },
        'Log': {
            'Node': 'with_node'
        },
        'Comment': {
            'Node': 'with_node',
            'User': 'with_user'
        }
    }

    all_fields_info, _ = get_all_fields_info()

    entity_prop = all_fields_info[current_entity_str].keys()

    project_cols = ['id']
    for prop in entity_prop:
        nprop = prop
        if current_entity_str in file_fields_to_model_fields:
            if prop in file_fields_to_model_fields[current_entity_str]:
                nprop = file_fields_to_model_fields[current_entity_str][prop]

        project_cols.append(nprop)

    # Here we should reference the entity of the main query
    current_entity_mod = entity_names_to_entities[current_entity_str]

    rel_string = relationship_dic[current_entity_str][originating_entity_str]
    mydict = {rel_string: entity_separator.join(tag_suffixes)}

    partial_query.append(
        current_entity_mod,
        tag=entity_separator.join(tag_suffixes) + entity_separator + current_entity_str,
        project=project_cols,
        outerjoin=True,
        **mydict
    )

    # prepare the recursion for the referenced entities
    foreign_fields = {
        k: v
        for k, v in all_fields_info[current_entity_str].items()
        # all_fields_info[model_name].items()
        if 'requires' in v
    }

    new_tag_suffixes = tag_suffixes + [current_entity_str]
    for value in foreign_fields.values():
        ref_model_name = value['requires']
        fill_in_query(partial_query, current_entity_str, ref_model_name, new_tag_suffixes)


def check_licenses(node_licenses, allowed_licenses, forbidden_licenses):
    """Check licenses"""
    from aiida.common.exceptions import LicensingException
    from inspect import isfunction

    for pk, license_ in node_licenses:
        if allowed_licenses is not None:
            try:
                if isfunction(allowed_licenses):
                    try:
                        if not allowed_licenses(license_):
                            raise LicensingException
                    except Exception:
                        raise LicensingException
                else:
                    if license_ not in allowed_licenses:
                        raise LicensingException
            except LicensingException:
                raise LicensingException(
                    f'Node {pk} is licensed under {license_} license, which is not in the list of allowed licenses'
                )
        if forbidden_licenses is not None:
            try:
                if isfunction(forbidden_licenses):
                    try:
                        if forbidden_licenses(license_):
                            raise LicensingException
                    except Exception:
                        raise LicensingException
                else:
                    if license_ in forbidden_licenses:
                        raise LicensingException
            except LicensingException:
                raise LicensingException(
                    f'Node {pk} is licensed under {license_} license, which is in the list of forbidden licenses'
                )


def serialize_field(data, track_conversion=False):
    """
    Serialize a single field.

    :todo: Generalize such that it the proper function is selected also during
        import
    """
    import datetime
    import pytz
    from uuid import UUID

    if isinstance(data, dict):
        ret_data = {}
        if track_conversion:
            ret_conversion = {}
            for key, value in data.items():
                ret_data[key], ret_conversion[key] = serialize_field(data=value, track_conversion=track_conversion)
        else:
            for key, value in data.items():
                ret_data[key] = serialize_field(data=value, track_conversion=track_conversion)
    elif isinstance(data, (list, tuple)):
        ret_data = []
        if track_conversion:
            ret_conversion = []
            for value in data:
                this_data, this_conversion = serialize_field(data=value, track_conversion=track_conversion)
                ret_data.append(this_data)
                ret_conversion.append(this_conversion)
        else:
            for value in data:
                ret_data.append(serialize_field(data=value, track_conversion=track_conversion))
    elif isinstance(data, datetime.datetime):
        # Note: requires timezone-aware objects!
        ret_data = data.astimezone(pytz.utc).strftime('%Y-%m-%dT%H:%M:%S.%f')
        ret_conversion = 'date'
    elif isinstance(data, UUID):
        ret_data = str(data)
        ret_conversion = None
    else:
        ret_data = data
        ret_conversion = None

    if track_conversion:
        return (ret_data, ret_conversion)
    return ret_data


def serialize_dict(datadict, remove_fields=None, rename_fields=None, track_conversion=False):
    """
    Serialize the dict using the serialize_field function to serialize
    each field.

    :param remove_fields: a list of strings.
      If a field with key inside the remove_fields list is found,
      it is removed from the dict.

      This is only used at level-0, no removal
      is possible at deeper levels.

    :param rename_fields: a dictionary in the format
      ``{"oldname": "newname"}``.

      If the "oldname" key is found, it is replaced with the
      "newname" string in the output dictionary.

      This is only used at level-0, no renaming
      is possible at deeper levels.
    :param track_conversion: if True, a tuple is returned, where the first
      element is the serialized dictionary, and the second element is a
      dictionary with the information on the serialized fields.
    """
    ret_dict = {}
    conversions = {}

    if not remove_fields:
        remove_fields = []

    if not rename_fields:
        rename_fields = {}

    for key, value in datadict.items():
        if key not in remove_fields:
            # rename_fields.get(key,key): use the replacement if found in rename_fields,
            # otherwise use 'key' as the default value.
            if track_conversion:
                (ret_dict[rename_fields.get(key, key)],
                 conversions[rename_fields.get(key,
                                               key)]) = serialize_field(data=value, track_conversion=track_conversion)
            else:
                ret_dict[rename_fields.get(key, key)] = serialize_field(data=value, track_conversion=track_conversion)

    if track_conversion:
        return (ret_dict, conversions)
    return ret_dict


def check_process_nodes_sealed(nodes):
    """Check ``ProcessNode`` s are sealed
    Only sealed ``ProcessNode`` s may be exported.

    :param nodes: :py:class:`~aiida.orm.nodes.process.process.ProcessNode` s to be checked. Should be their PK(s).
    :type nodes: list, int

    :raises `~aiida.tools.importexport.common.exceptions.ExportValidationError`:
        if a ``ProcessNode`` is not sealed or `nodes` is not a `list`, `set`, or `int`.
    """
    if not nodes:
        return

    # Check `nodes` type, and if necessary change to set
    if isinstance(nodes, set):
        pass
    elif isinstance(nodes, list):
        nodes = set(nodes)
    elif isinstance(nodes, int):
        nodes = set([nodes])
    else:
        raise exceptions.ExportValidationError('nodes must be either an int or set/list of ints')

    filters = {'id': {'in': nodes}, 'attributes.sealed': True}
    sealed_nodes = set(QueryBuilder().append(ProcessNode, filters=filters, project=['id']).all(flat=True))

    if sealed_nodes != nodes:
        raise exceptions.ExportValidationError(
            'All ProcessNodes must be sealed before they can be exported. '
            'Node(s) with PK(s): {} is/are not sealed.'.format(', '.join(str(pk) for pk in nodes - sealed_nodes))
        )


def summary(*, file_format, export_version, outfile, include_comments, include_logs, traversal_rules):
    """Print summary for export"""
    from tabulate import tabulate

    parameters = [['Archive', outfile], ['Format', file_format], ['Export version', export_version]]

    result = f"\n{tabulate(parameters, headers=['EXPORT', ''])}"

    inclusions = [['Include Comments', include_comments], ['Include Logs', include_logs]]
    result += f"\n\n{tabulate(inclusions, headers=['Inclusion rules', ''])}"

    traversal_rules = [[f"Follow links {' '.join(name.split('_'))}s", value] for name, value in traversal_rules.items()]
    result += f"\n\n{tabulate(traversal_rules, headers=['Traversal rules', ''])}\n"

    EXPORT_LOGGER.log(msg=result, level=LOG_LEVEL_REPORT)


def deprecated_parameters(old, new):
    """Handle deprecated parameter (where it is replaced with another)

    :param old: The old, deprecated parameter as a dict with keys "name" and "value"
    :type old: dict

    :param new: The new parameter as a dict with keys "name" and "value"
    :type new: dict

    :return: New parameter's value (if not defined, then old parameter's value)
    """
    if old.get('value', None) is not None:
        if new.get('value', None) is not None:
            message = f"`{old['name']}` is deprecated, the supplied `{new['name']}` input will be used"
        else:
            message = f"`{old['name']}` is deprecated, please use `{new['name']}` instead"
            new['value'] = old['value']
        warnings.warn(message, AiidaDeprecationWarning)  # pylint: disable=no-member

    return new['value']



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Data class that can be used to store a single file in its repository."""
import inspect
import os
import warnings

from aiida.common import exceptions
from aiida.common.warnings import AiidaDeprecationWarning
from .data import Data

__all__ = ('SinglefileData',)


class SinglefileData(Data):
    """Data class that can be used to store a single file in its repository."""

    DEFAULT_FILENAME = 'file.txt'

    def __init__(self, file, filename=None, **kwargs):
        """Construct a new instance and set the contents to that of the file.

        :param file: an absolute filepath or filelike object whose contents to copy.
            Hint: Pass io.BytesIO(b"my string") to construct the SinglefileData directly from a string.
        :param filename: specify filename to use (defaults to name of provided file).
        """
        # pylint: disable=redefined-builtin
        super().__init__(**kwargs)

        # 'filename' argument was added to 'set_file' after 1.0.0.
        if 'filename' not in inspect.getfullargspec(self.set_file)[0]:
            warnings.warn(  # pylint: disable=no-member
                f"Method '{type(self).__name__}.set_file' does not support the 'filename' argument. " +
                'This will raise an exception in AiiDA 2.0.', AiidaDeprecationWarning
            )

        if file is not None:
            if filename is None:
                # don't assume that set_file has a 'filename' argument (remove guard in 2.0.0)
                self.set_file(file)
            else:
                self.set_file(file, filename=filename)

    @property
    def filename(self):
        """Return the name of the file stored.

        :return: the filename under which the file is stored in the repository
        """
        return self.get_attribute('filename')

    def open(self, path=None, mode='r', key=None):
        """Return an open file handle to the content of this data node.

        .. deprecated:: 1.4.0
            Keyword `key` is deprecated and will be removed in `v2.0.0`. Use `path` instead.

        .. deprecated:: 1.4.0
            Starting from `v2.0.0` this will raise if not used in a context manager.

        :param path: the relative path of the object within the repository.
        :param key: optional key within the repository, by default is the `filename` set in the attributes
        :param mode: the mode with which to open the file handle (default: read mode)
        :return: a file handle
        """
        from ..node import WarnWhenNotEntered
        if key is not None:
            if path is not None:
                raise ValueError('cannot specify both `path` and `key`.')
            warnings.warn(
                'keyword `key` is deprecated and will be removed in `v2.0.0`. Use `path` instead.',
                AiidaDeprecationWarning
            )  # pylint: disable=no-member
            path = key

        if path is None:
            path = self.filename

        return WarnWhenNotEntered(self._repository.open(path, mode=mode), repr(self))

    def get_content(self):
        """Return the content of the single file stored for this data node.

        :return: the content of the file as a string
        """
        with self.open() as handle:
            return handle.read()

    def set_file(self, file, filename=None):
        """Store the content of the file in the node's repository, deleting any other existing objects.

        :param file: an absolute filepath or filelike object whose contents to copy
            Hint: Pass io.BytesIO(b"my string") to construct the file directly from a string.
        :param filename: specify filename to use (defaults to name of provided file).
        """
        # pylint: disable=redefined-builtin

        if isinstance(file, str):
            is_filelike = False

            key = os.path.basename(file)
            if not os.path.isabs(file):
                raise ValueError(f'path `{file}` is not absolute')

            if not os.path.isfile(file):
                raise ValueError(f'path `{file}` does not correspond to an existing file')
        else:
            is_filelike = True
            try:
                key = os.path.basename(file.name)
            except AttributeError:
                key = self.DEFAULT_FILENAME

        key = filename or key

        existing_object_names = self.list_object_names()

        try:
            # Remove the 'key' from the list of currently existing objects such that it is not deleted after storing
            existing_object_names.remove(key)
        except ValueError:
            pass

        if is_filelike:
            self.put_object_from_filelike(file, key, mode='wb')
        else:
            self.put_object_from_file(file, key)

        # Delete any other existing objects (minus the current `key` which was already removed from the list)
        for existing_key in existing_object_names:
            self.delete_object(existing_key)

        self.set_attribute('filename', key)

    def _validate(self):
        """Ensure that there is one object stored in the repository, whose key matches value set for `filename` attr."""
        super()._validate()

        try:
            filename = self.filename
        except AttributeError:
            raise exceptions.ValidationError('the `filename` attribute is not set.')

        objects = self.list_object_names()

        if [filename] != objects:
            raise exceptions.ValidationError(
                f'respository files {objects} do not match the `filename` attribute {filename}.'
            )



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Deleting dbpath table and triggers

Revision ID: 70c7d732f1b2
Revises:
Create Date: 2017-10-17 10:30:23.327195

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.orm.session import Session
from aiida.backends.sqlalchemy.utils import install_tc

# revision identifiers, used by Alembic.
revision = '70c7d732f1b2'
down_revision = 'e15ef2630a1b'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    op.drop_table('db_dbpath')
    conn = op.get_bind()
    conn.execute('DROP TRIGGER IF EXISTS autoupdate_tc ON db_dblink')
    conn.execute('DROP FUNCTION IF EXISTS update_tc()')


def downgrade():
    """Migrations for the downgrade."""
    op.create_table(
        'db_dbpath', sa.Column('id', sa.INTEGER(), nullable=False),
        sa.Column('parent_id', sa.INTEGER(), autoincrement=False, nullable=True),
        sa.Column('child_id', sa.INTEGER(), autoincrement=False, nullable=True),
        sa.Column('depth', sa.INTEGER(), autoincrement=False, nullable=True),
        sa.Column('entry_edge_id', sa.INTEGER(), autoincrement=False, nullable=True),
        sa.Column('direct_edge_id', sa.INTEGER(), autoincrement=False, nullable=True),
        sa.Column('exit_edge_id', sa.INTEGER(), autoincrement=False, nullable=True),
        sa.ForeignKeyConstraint(['child_id'], ['db_dbnode.id'],
                                name='db_dbpath_child_id_fkey',
                                initially='DEFERRED',
                                deferrable=True),
        sa.ForeignKeyConstraint(['parent_id'], ['db_dbnode.id'],
                                name='db_dbpath_parent_id_fkey',
                                initially='DEFERRED',
                                deferrable=True), sa.PrimaryKeyConstraint('id', name='db_dbpath_pkey')
    )
    # I get the session using the alembic connection
    # (Keep in mind that alembic uses the AiiDA SQLA
    # session)
    session = Session(bind=op.get_bind())
    install_tc(session)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""SQLA groups"""

import collections
import logging

from aiida.backends import sqlalchemy as sa
from aiida.backends.sqlalchemy.models.group import DbGroup, table_groups_nodes
from aiida.backends.sqlalchemy.models.node import DbNode
from aiida.common.exceptions import UniquenessError
from aiida.common.lang import type_check
from aiida.orm.implementation.groups import BackendGroup, BackendGroupCollection
from . import entities
from . import users
from . import utils

__all__ = ('SqlaGroup', 'SqlaGroupCollection')

_LOGGER = logging.getLogger(__name__)


# Unfortunately the linter doesn't seem to be able to pick up on the fact that the abstract property 'id'
# of BackendGroup is actually implemented in SqlaModelEntity so disable the abstract check
class SqlaGroup(entities.SqlaModelEntity[DbGroup], BackendGroup):  # pylint: disable=abstract-method
    """The SQLAlchemy Group object"""

    MODEL_CLASS = DbGroup

    def __init__(self, backend, label, user, description='', type_string=''):
        """
        Construct a new SQLA group

        :param backend: the backend to use
        :param label: the group label
        :param user: the owner of the group
        :param description: an optional group description
        :param type_string: an optional type for the group to contain
        """
        type_check(user, users.SqlaUser)
        super().__init__(backend)

        dbgroup = DbGroup(label=label, description=description, user=user.dbmodel, type_string=type_string)
        self._dbmodel = utils.ModelWrapper(dbgroup)

    @property
    def label(self):
        return self._dbmodel.label

    @label.setter
    def label(self, label):
        """
        Attempt to change the label of the group instance. If the group is already stored
        and the another group of the same type already exists with the desired label, a
        UniquenessError will be raised

        :param label: the new group label
        :raises aiida.common.UniquenessError: if another group of same type and label already exists
        """
        self._dbmodel.label = label

        if self.is_stored:
            try:
                self._dbmodel.save()
            except Exception:
                raise UniquenessError(f'a group of the same type with the label {label} already exists') \
                    from Exception

    @property
    def description(self):
        return self._dbmodel.description

    @description.setter
    def description(self, value):
        self._dbmodel.description = value

        # Update the entry in the DB, if the group is already stored
        if self.is_stored:
            self._dbmodel.save()

    @property
    def type_string(self):
        return self._dbmodel.type_string

    @property
    def user(self):
        return self._backend.users.from_dbmodel(self._dbmodel.user)

    @user.setter
    def user(self, new_user):
        type_check(new_user, users.SqlaUser)
        self._dbmodel.user = new_user.dbmodel

    @property
    def pk(self):
        return self._dbmodel.id

    @property
    def uuid(self):
        return str(self._dbmodel.uuid)

    def __int__(self):
        if not self.is_stored:
            return None

        return self._dbnode.id  # pylint: disable=no-member

    @property
    def is_stored(self):
        return self.pk is not None

    def store(self):
        self._dbmodel.save()
        return self

    def count(self):
        """Return the number of entities in this group.

        :return: integer number of entities contained within the group
        """
        from aiida.backends.sqlalchemy import get_scoped_session
        session = get_scoped_session()
        return session.query(self.MODEL_CLASS).join(self.MODEL_CLASS.dbnodes).filter(DbGroup.id == self.pk).count()

    def clear(self):
        """Remove all the nodes from this group."""
        from aiida.backends.sqlalchemy import get_scoped_session
        session = get_scoped_session()
        # Note we have to call `dbmodel` and `_dbmodel` to circumvent the `ModelWrapper`
        self.dbmodel.dbnodes = []
        session.commit()

    @property
    def nodes(self):
        """Get an iterator to all the nodes in the group"""

        class Iterator:
            """Nodes iterator"""

            def __init__(self, dbnodes, backend):
                self._backend = backend
                self._dbnodes = dbnodes
                self.generator = self._genfunction()

            def _genfunction(self):
                for node in self._dbnodes:
                    yield self._backend.get_backend_entity(node)

            def __iter__(self):
                return self

            def __len__(self):
                return self._dbnodes.count()

            def __getitem__(self, value):
                if isinstance(value, slice):
                    return [self._backend.get_backend_entity(n) for n in self._dbnodes[value]]

                return self._backend.get_backend_entity(self._dbnodes[value])

            def __next__(self):
                return next(self.generator)

        return Iterator(self._dbmodel.dbnodes, self._backend)

    def add_nodes(self, nodes, **kwargs):
        """Add a node or a set of nodes to the group.

        :note: all the nodes *and* the group itself have to be stored.

        :param nodes: a list of `BackendNode` instance to be added to this group

        :param kwargs:
            skip_orm: When the flag is on, the SQLA ORM is skipped and SQLA is used
            to create a direct SQL INSERT statement to the group-node relationship
            table (to improve speed).
        """
        from sqlalchemy.exc import IntegrityError  # pylint: disable=import-error, no-name-in-module
        from sqlalchemy.dialects.postgresql import insert  # pylint: disable=import-error, no-name-in-module
        from aiida.orm.implementation.sqlalchemy.nodes import SqlaNode
        from aiida.backends.sqlalchemy import get_scoped_session
        from aiida.backends.sqlalchemy.models.base import Base

        super().add_nodes(nodes)
        skip_orm = kwargs.get('skip_orm', False)

        def check_node(given_node):
            """ Check if given node is of correct type and stored """
            if not isinstance(given_node, SqlaNode):
                raise TypeError(f'invalid type {type(given_node)}, has to be {SqlaNode}')

            if not given_node.is_stored:
                raise ValueError('At least one of the provided nodes is unstored, stopping...')

        with utils.disable_expire_on_commit(get_scoped_session()) as session:
            if not skip_orm:
                # Get dbnodes here ONCE, otherwise each call to dbnodes will re-read the current value in the database
                dbnodes = self._dbmodel.dbnodes

                for node in nodes:
                    check_node(node)

                    # Use pattern as suggested here:
                    # http://docs.sqlalchemy.org/en/latest/orm/session_transaction.html#using-savepoint
                    try:
                        with session.begin_nested():
                            dbnodes.append(node.dbmodel)
                            session.flush()
                    except IntegrityError:
                        # Duplicate entry, skip
                        pass
            else:
                ins_dict = list()
                for node in nodes:
                    check_node(node)
                    ins_dict.append({'dbnode_id': node.id, 'dbgroup_id': self.id})

                my_table = Base.metadata.tables['db_dbgroup_dbnodes']
                ins = insert(my_table).values(ins_dict)
                session.execute(ins.on_conflict_do_nothing(index_elements=['dbnode_id', 'dbgroup_id']))

            # Commit everything as up till now we've just flushed
            session.commit()

    def remove_nodes(self, nodes, **kwargs):
        """Remove a node or a set of nodes from the group.

        :note: all the nodes *and* the group itself have to be stored.

        :param nodes: a list of `BackendNode` instance to be added to this group
        :param kwargs:
            skip_orm: When the flag is set to `True`, the SQLA ORM is skipped and SQLA is used to create a direct SQL
            DELETE statement to the group-node relationship table in order to improve speed.
        """
        from sqlalchemy import and_
        from aiida.backends.sqlalchemy import get_scoped_session
        from aiida.backends.sqlalchemy.models.base import Base
        from aiida.orm.implementation.sqlalchemy.nodes import SqlaNode

        super().remove_nodes(nodes)

        # Get dbnodes here ONCE, otherwise each call to dbnodes will re-read the current value in the database
        dbnodes = self._dbmodel.dbnodes
        skip_orm = kwargs.get('skip_orm', False)

        def check_node(node):
            if not isinstance(node, SqlaNode):
                raise TypeError(f'invalid type {type(node)}, has to be {SqlaNode}')

            if node.id is None:
                raise ValueError('At least one of the provided nodes is unstored, stopping...')

        list_nodes = []

        with utils.disable_expire_on_commit(get_scoped_session()) as session:
            if not skip_orm:
                for node in nodes:
                    check_node(node)

                    # Check first, if SqlA issues a DELETE statement for an unexisting key it will result in an error
                    if node.dbmodel in dbnodes:
                        list_nodes.append(node.dbmodel)

                for node in list_nodes:
                    dbnodes.remove(node)
            else:
                table = Base.metadata.tables['db_dbgroup_dbnodes']
                for node in nodes:
                    check_node(node)
                    clause = and_(table.c.dbnode_id == node.id, table.c.dbgroup_id == self.id)
                    statement = table.delete().where(clause)
                    session.execute(statement)

            session.commit()


class SqlaGroupCollection(BackendGroupCollection):
    """The SLQA collection of groups"""

    ENTITY_CLASS = SqlaGroup

    def query(
        self,
        label=None,
        type_string=None,
        pk=None,
        uuid=None,
        nodes=None,
        user=None,
        node_attributes=None,
        past_days=None,
        label_filters=None,
        **kwargs
    ):  # pylint: disable=too-many-arguments
        # pylint: disable=too-many-branches
        from aiida.orm.implementation.sqlalchemy.nodes import SqlaNode

        session = sa.get_scoped_session()

        filters = []

        if label is not None:
            filters.append(DbGroup.label == label)
        if type_string is not None:
            filters.append(DbGroup.type_string == type_string)
        if pk is not None:
            filters.append(DbGroup.id == pk)
        if uuid is not None:
            filters.append(DbGroup.uuid == uuid)
        if past_days is not None:
            filters.append(DbGroup.time >= past_days)
        if nodes:
            if not isinstance(nodes, collections.Iterable):
                nodes = [nodes]

            if not all(isinstance(n, (SqlaNode, DbNode)) for n in nodes):
                raise TypeError(
                    'At least one of the elements passed as '
                    'nodes for the query on Group is neither '
                    'a Node nor a DbNode'
                )

            # In the case of the Node orm from Sqlalchemy, there is an id
            # property on it.
            sub_query = (
                session.query(table_groups_nodes).filter(
                    table_groups_nodes.c['dbnode_id'].in_([n.id for n in nodes]),
                    table_groups_nodes.c['dbgroup_id'] == DbGroup.id
                ).exists()
            )

            filters.append(sub_query)
        if user:
            if isinstance(user, str):
                filters.append(DbGroup.user.has(email=user.email))
            else:
                type_check(user, users.SqlaUser)
                filters.append(DbGroup.user == user.dbmodel)

        if label_filters:
            for key, value in label_filters.items():
                if not value:
                    continue
                if key == 'startswith':
                    filters.append(DbGroup.label.like(f'{value}%'))
                elif key == 'endswith':
                    filters.append(DbGroup.label.like(f'%{value}'))
                elif key == 'contains':
                    filters.append(DbGroup.label.like(f'%{value}%'))

        if node_attributes:
            _LOGGER.warning("SQLA query doesn't support node attribute filters, ignoring '%s'", node_attributes)

        if kwargs:
            _LOGGER.warning("SQLA query doesn't support additional filters, ignoring '%s'", kwargs)
        groups = (session.query(DbGroup).filter(*filters).order_by(DbGroup.id).distinct().all())

        return [SqlaGroup.from_dbmodel(group, self._backend) for group in groups]

    def delete(self, id):  # pylint: disable=redefined-builtin
        session = sa.get_scoped_session()

        session.query(DbGroup).get(id).delete()
        session.commit()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Shared resources for the archive."""
from collections import OrderedDict
import copy
import dataclasses
import os
from pathlib import Path
import tarfile
from types import TracebackType
from typing import Any, Dict, List, Optional, Tuple, Type, Union
import zipfile

from aiida.common import json  # handles byte dumps
from aiida.common.log import AIIDA_LOGGER

__all__ = ('ArchiveMetadata', 'detect_archive_type', 'null_callback', 'CacheFolder')

ARCHIVE_LOGGER = AIIDA_LOGGER.getChild('archive')


@dataclasses.dataclass
class ArchiveMetadata:
    """Class for storing metadata about this archive.

    Required fields are necessary for importing the data back into AiiDA,
    whereas optional fields capture information about the export/migration process(es)
    """
    export_version: str
    aiida_version: str
    # Entity type -> database ID key
    unique_identifiers: Dict[str, str] = dataclasses.field(repr=False)
    # Entity type -> database key -> meta parameters
    all_fields_info: Dict[str, Dict[str, Dict[str, str]]] = dataclasses.field(repr=False)

    # optional data
    graph_traversal_rules: Optional[Dict[str, bool]] = dataclasses.field(default=None)
    # Entity type -> UUID list
    entities_starting_set: Optional[Dict[str, List[str]]] = dataclasses.field(default=None)
    include_comments: Optional[bool] = dataclasses.field(default=None)
    include_logs: Optional[bool] = dataclasses.field(default=None)
    # list of migration event notifications
    conversion_info: List[str] = dataclasses.field(default_factory=list, repr=False)


def null_callback(action: str, value: Any):  # pylint: disable=unused-argument
    """A null callback function."""


def detect_archive_type(in_path: str) -> str:
    """For back-compatibility, but should be replaced with direct comparison of classes.

    :param in_path: the path to the file
    :returns: the archive type identifier (currently one of 'zip', 'tar.gz', 'folder')

    """
    from aiida.tools.importexport.common.config import ExportFileFormat
    from aiida.tools.importexport.common.exceptions import ImportValidationError

    if os.path.isdir(in_path):
        return 'folder'
    if tarfile.is_tarfile(in_path):
        return ExportFileFormat.TAR_GZIPPED
    if zipfile.is_zipfile(in_path):
        return ExportFileFormat.ZIP
    raise ImportValidationError(
        'Unable to detect the input file format, it is neither a '
        'folder, tar file, nor a (possibly compressed) zip file.'
    )


class CacheFolder:
    """A class to encapsulate a folder path with cached read/writes.

    The class can be used as a context manager, and will flush the cache on exit::

        with CacheFolder(path) as folder:
            # these are stored in memory (no disk write)
            folder.write_text('path/to/file.txt', 'content')
            folder.write_json('path/to/data.json', {'a': 1})
            # these will be read from memory
            text = folder.read_text('path/to/file.txt')
            text = folder.load_json('path/to/data.json')

        # all files will now have been written to disk

    """

    def __init__(self, path: Union[Path, str], *, encoding: str = 'utf8'):
        """Initialise cached folder.

        :param path: folder path to cache
        :param encoding: encoding of text to read/write

        """
        self._path = Path(path)
        # dict mapping path -> (type, content)
        self._cache = OrderedDict()  # type: ignore
        self._encoding = encoding
        self._max_items = 100  # maximum limit of files to store in memory

    def _write_object(self, path: str, ctype: str, content: Any):
        """Write an object from the cache to disk.

        :param path: relative path of file
        :param ctype: the type of the content
        :param content: the content to write

        """
        if ctype == 'text':
            (self._path / path).write_text(content, encoding=self._encoding)
        elif ctype == 'json':
            with (self._path / path).open(mode='wb') as handle:
                json.dump(content, handle)
        else:
            raise TypeError(f'Unknown content type: {ctype}')

    def flush(self):
        """Flush the cache."""
        for path, (ctype, content) in self._cache.items():
            self._write_object(path, ctype, content)

    def _limit_cache(self):
        """Ensure the cache does not exceed a set limit.

        Content is uncached on a First-In-First-Out basis.

        """
        while len(self._cache) > self._max_items:
            path, (ctype, content) = self._cache.popitem(last=False)
            self._write_object(path, ctype, content)

    def get_path(self, flush=True) -> Path:
        """Return the path.

        :param flush: flush the cache before returning

        """
        if flush:
            self.flush()
        return self._path

    def write_text(self, path: str, content: str):
        """write text to the cache.

        :param path: path relative to base folder

        """
        assert isinstance(content, str)
        self._cache[path] = ('text', content)
        self._limit_cache()

    def read_text(self, path) -> str:
        """write text from the cache or base folder.

        :param path: path relative to base folder

        """
        if path not in self._cache:
            return (self._path / path).read_text(self._encoding)
        ctype, content = self._cache[path]
        if ctype == 'text':
            return content
        if ctype == 'json':
            return json.dumps(content)

        raise TypeError(f"content of type '{ctype}' could not be converted to text")

    def write_json(self, path: str, data: dict):
        """Write dict to the folder, to be serialized as json.

        The dictionary is stored in memory, until the cache is flushed,
        at which point the dictionary is serialized to json and written to disk.

        :param path: path relative to base folder

        """
        assert isinstance(data, dict)
        # json.dumps(data)  # make sure that the data can be converted to json (increases memory usage)
        self._cache[path] = ('json', data)
        self._limit_cache()

    def load_json(self, path: str, ensure_copy: bool = False) -> Tuple[bool, dict]:
        """Load a json file from the cache folder.

        Important: if the dict is returned directly from the cache, any mutations will affect the cached dict.

        :param path: path relative to base folder
        :param ensure_copy: ensure the dict is a copy of that from the cache

        :returns: (from cache, the content)
            If from cache, mutations will directly affect the cache

        """
        if path not in self._cache:
            return False, json.loads((self._path / path).read_text(self._encoding))

        ctype, content = self._cache[path]
        if ctype == 'text':
            return False, json.loads(content)
        if ctype == 'json':
            if ensure_copy:
                return False, copy.deepcopy(content)
            return True, content

        raise TypeError(f"content of type '{ctype}' could not be converted to a dict")

    def remove_file(self, path):
        """Remove a file from both the cache and base folder (if present).

        :param path: path relative to base folder

        """
        self._cache.pop(path, None)
        if (self._path / path).exists():
            (self._path / path).unlink()

    def __enter__(self):
        """Enter the contextmanager."""
        return self

    def __exit__(
        self, exctype: Optional[Type[BaseException]], excinst: Optional[BaseException], exctb: Optional[TracebackType]
    ):
        """Exit the contextmanager."""
        self.flush()
        return False



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Utilities that extend the basic python language."""
import functools
import inspect
import keyword


def isidentifier(identifier):
    """Return whether the given string is a valid python identifier.

    :return: boolean, True if identifier is valid, False otherwise
    :raises TypeError: if identifier is not string type
    """
    type_check(identifier, str)
    return identifier.isidentifier() and not keyword.iskeyword(identifier)


def type_check(what, of_type, msg=None, allow_none=False):
    """Verify that object 'what' is of type 'of_type' and if not the case, raise a TypeError.

    :param what: the object to check
    :param of_type: the type (or tuple of types) to compare to
    :param msg: if specified, allows to customize the message that is passed within the TypeError exception
    :param allow_none: boolean, if True will not raise if the passed `what` is `None`

    :return: `what` or `None`
    """
    if allow_none and what is None:
        return None

    if not isinstance(what, of_type):
        if msg is None:
            msg = f"Got object of type '{type(what)}', expecting '{of_type}'"
        raise TypeError(msg)

    return what


def override_decorator(check=False):
    """Decorator to signal that a method from a base class is being overridden completely."""

    def wrap(func):  # pylint: disable=missing-docstring
        if isinstance(func, property):
            raise RuntimeError('Override must go after @property decorator')

        args = inspect.getfullargspec(func)[0]
        if not args:
            raise RuntimeError('Can only use the override decorator on member functions')

        if check:

            @functools.wraps(func)
            def wrapped_fn(self, *args, **kwargs):  # pylint: disable=missing-docstring
                try:
                    getattr(super(), func.__name__)
                except AttributeError:
                    raise RuntimeError(f'Function {func} does not override a superclass method')

                return func(self, *args, **kwargs)
        else:
            wrapped_fn = func

        return wrapped_fn

    return wrap


override = override_decorator(check=False)  # pylint: disable=invalid-name


class classproperty:  # pylint: disable=invalid-name
    """
    A class that, when used as a decorator, works as if the
    two decorators @property and @classmethod where applied together
    (i.e., the object works as a property, both for the Class and for any
    of its instance; and is called with the class cls rather than with the
    instance as its first argument).
    """

    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(owner)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Function to delete nodes from the database."""

import click
from aiida.cmdline.utils import echo


def delete_nodes(pks, verbosity=0, dry_run=False, force=False, **kwargs):
    """Delete nodes by a list of pks.

    This command will delete not only the specified nodes, but also the ones that are
    linked to these and should be also deleted in order to keep a consistent provenance
    according to the rules explained in the concepts section of the documentation.
    In summary:

    1. If a DATA node is deleted, any process nodes linked to it will also be deleted.

    2. If a CALC node is deleted, any incoming WORK node (callers) will be deleted as
    well whereas any incoming DATA node (inputs) will be kept. Outgoing DATA nodes
    (outputs) will be deleted by default but this can be disabled.

    3. If a WORK node is deleted, any incoming WORK node (callers) will be deleted as
    well, but all DATA nodes will be kept. Outgoing WORK or CALC nodes will be kept by
    default, but deletion of either of both kind of connected nodes can be enabled.

    These rules are 'recursive', so if a CALC node is deleted, then its output DATA
    nodes will be deleted as well, and then any CALC node that may have those as
    inputs, and so on.

    :param pks: a list of the PKs of the nodes to delete
    :param bool force: do not ask for confirmation to delete nodes.
    :param int verbosity: 0 prints nothing,
                          1 prints just sums and total,
                          2 prints individual nodes.

    :param kwargs: graph traversal rules. See :const:`aiida.common.links.GraphTraversalRules` what rule names
        are toggleable and what the defaults are.
    :param bool dry_run:
        Just perform a dry run and do not delete anything. Print statistics according
        to the verbosity level set.
    :param bool force:
        Do not ask for confirmation to delete nodes.
    """
    # pylint: disable=too-many-arguments,too-many-branches,too-many-locals,too-many-statements
    from aiida.backends.utils import delete_nodes_and_connections
    from aiida.common import exceptions
    from aiida.orm import Node, QueryBuilder, load_node
    from aiida.tools.graph.graph_traversers import get_nodes_delete

    starting_pks = []
    for pk in pks:
        try:
            load_node(pk)
        except exceptions.NotExistent:
            echo.echo_warning(f'warning: node with pk<{pk}> does not exist, skipping')
        else:
            starting_pks.append(pk)

    # An empty set might be problematic for the queries done below.
    if not starting_pks:
        if verbosity:
            echo.echo('Nothing to delete')
        return

    pks_set_to_delete = get_nodes_delete(starting_pks, **kwargs)['nodes']

    if verbosity > 0:
        echo.echo(
            'I {} delete {} node{}'.format(
                'would' if dry_run else 'will', len(pks_set_to_delete), 's' if len(pks_set_to_delete) > 1 else ''
            )
        )
        if verbosity > 1:
            builder = QueryBuilder().append(
                Node, filters={'id': {
                    'in': pks_set_to_delete
                }}, project=('uuid', 'id', 'node_type', 'label')
            )
            echo.echo(f"The nodes I {'would' if dry_run else 'will'} delete:")
            for uuid, pk, type_string, label in builder.iterall():
                try:
                    short_type_string = type_string.split('.')[-2]
                except IndexError:
                    short_type_string = type_string
                echo.echo(f'   {uuid} {pk} {short_type_string} {label}')

    if dry_run:
        if verbosity > 0:
            echo.echo('\nThis was a dry run, exiting without deleting anything')
        return

    # Asking for user confirmation here
    if force:
        pass
    else:
        echo.echo_warning(f'YOU ARE ABOUT TO DELETE {len(pks_set_to_delete)} NODES! THIS CANNOT BE UNDONE!')
        if not click.confirm('Shall I continue?'):
            echo.echo('Exiting without deleting')
            return

    # Recover the list of folders to delete before actually deleting the nodes. I will delete the folders only later,
    # so that if there is a problem during the deletion of the nodes in the DB, I don't delete the folders
    repositories = [load_node(pk)._repository for pk in pks_set_to_delete]  # pylint: disable=protected-access

    if verbosity > 0:
        echo.echo('Starting node deletion...')
    delete_nodes_and_connections(pks_set_to_delete)

    if verbosity > 0:
        echo.echo('Nodes deleted from database, deleting files from the repository now...')

    # If we are here, we managed to delete the entries from the DB.
    # I can now delete the folders
    for repository in repositories:
        repository.erase(force=True)

    if verbosity > 0:
        echo.echo('Deletion completed.')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Testing infrastructure for easy testing of AiiDA plugins.

"""

import warnings
from aiida.common.warnings import AiidaDeprecationWarning
from aiida.manage.tests import TestManager as FixtureManager
from aiida.manage.tests import test_manager as fixture_manager
from aiida.manage.tests import _GLOBAL_TEST_MANAGER as _GLOBAL_FIXTURE_MANAGER
from aiida.manage.tests.unittest_classes import PluginTestCase

warnings.warn('this module is deprecated, use `aiida.manage.tests` and its submodules instead', AiidaDeprecationWarning)  # pylint: disable=no-member

__all__ = ('FixtureManager', 'fixture_manager', '_GLOBAL_FIXTURE_MANAGER', 'PluginTestCase')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Translator for data node
"""

from aiida.restapi.translator.nodes.node import NodeTranslator
from aiida.restapi.common.exceptions import RestInputValidationError
from aiida.common.exceptions import LicensingException


class DataTranslator(NodeTranslator):
    """
    Translator relative to resource 'data' and aiida class `~aiida.orm.nodes.data.data.Data`
    """

    # A label associated to the present class (coincides with the resource name)
    __label__ = 'data'
    # The AiiDA class one-to-one associated to the present class
    from aiida.orm import Data
    _aiida_class = Data
    # The string name of the AiiDA class
    _aiida_type = 'data.Data'

    _result_type = __label__

    @staticmethod
    def get_downloadable_data(node, download_format=None):
        """
        Return content in the specified format for download

        :param node: node representing cif file to be downloaded
        :param download_format: export format
        :returns: content of the node in the specified format for download
        """
        response = {}

        if download_format is None:
            raise RestInputValidationError(
                'Please specify the download format. '
                'The available download formats can be '
                'queried using the /nodes/download_formats/ endpoint.'
            )

        elif download_format in node.get_export_formats():
            try:
                response['data'] = node._exportcontent(download_format)[0]  # pylint: disable=protected-access
                response['status'] = 200
                try:
                    response['filename'] = node.filename
                except AttributeError:
                    response['filename'] = f'{node.uuid}.{download_format}'
            except LicensingException as exc:
                response['status'] = 500
                response['data'] = str(exc)

            return response

        else:
            raise RestInputValidationError(
                'The format {} is not supported. '
                'The available download formats can be '
                'queried using the /nodes/download_formats/ endpoint.'.format(download_format)
            )



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module with the implementations of the various backend entities for various database backends."""
# pylint: disable=wildcard-import,undefined-variable
from .authinfos import *
from .backends import *
from .comments import *
from .computers import *
from .groups import *
from .logs import *
from .nodes import *
from .querybuilder import *
from .users import *

__all__ = (
    authinfos.__all__ + backends.__all__ + comments.__all__ + computers.__all__ + groups.__all__ + logs.__all__ +
    nodes.__all__ + querybuilder.__all__ + users.__all__
)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=import-error,no-name-in-module
"""Module to manage users for the SQLA backend."""

from sqlalchemy.schema import Column
from sqlalchemy.types import Integer, String

from aiida.backends.sqlalchemy.models.base import Base


class DbUser(Base):
    """Store users using the SQLA backend."""
    __tablename__ = 'db_dbuser'

    id = Column(Integer, primary_key=True)  # pylint: disable=invalid-name
    email = Column(String(254), unique=True, index=True)
    first_name = Column(String(254), nullable=True)
    last_name = Column(String(254), nullable=True)
    institution = Column(String(254), nullable=True)

    def __init__(self, email, first_name='', last_name='', institution='', **kwargs):
        """Set additional class attributes with respect to the base class."""
        self.email = email
        self.first_name = first_name
        self.last_name = last_name
        self.institution = institution
        super().__init__(**kwargs)

    def __str__(self):
        return self.email



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Prepare schema reset.

Revision ID: 91b573400be5
Revises: 7b38a9e783e7
Create Date: 2019-07-25 14:58:39.866822

"""

# Remove when https://github.com/PyCQA/pylint/issues/1931 is fixed
# pylint: disable=no-name-in-module,import-error
from alembic import op
from sqlalchemy.sql import text

# revision identifiers, used by Alembic.
revision = '91b573400be5'
down_revision = '7b38a9e783e7'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    conn = op.get_bind()

    # The following statement is trying to perform an UPSERT, i.e. an UPDATE of a given key or if it doesn't exist fall
    # back to an INSERT. This problem is notoriously difficult to solve as explained in great detail in this article:
    # https://www.depesz.com/2012/06/10/why-is-upsert-so-complicated/ Postgres 9.5 provides an offical UPSERT method
    # through the `ON CONFLICT` keyword, but since we also support 9.4 we cannot use it here. The snippet used below
    # taken from the provided link, is not safe for concurrent operations, but since our migrations always run in an
    # isolated way, we do not suffer from those problems and can safely use it.
    statement = text(
        """
        INSERT INTO db_dbsetting (key, val, description)
        SELECT 'schema_generation', '"1"', 'Database schema generation'
        WHERE NOT EXISTS (SELECT * FROM db_dbsetting WHERE key = 'schema_generation');
        """
    )
    conn.execute(statement)


def downgrade():
    """Migrations for the downgrade."""



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=import-error,no-name-in-module
"""Module to manage computers for the SQLA backend."""

from sqlalchemy import ForeignKey
from sqlalchemy.orm import relationship, backref
from sqlalchemy.schema import Column, Table, UniqueConstraint, Index
from sqlalchemy.types import Integer, String, DateTime, Text

from sqlalchemy.dialects.postgresql import UUID, JSONB

from aiida.common import timezone
from aiida.common.utils import get_new_uuid

from .base import Base

table_groups_nodes = Table(  # pylint: disable=invalid-name
    'db_dbgroup_dbnodes',
    Base.metadata,
    Column('id', Integer, primary_key=True),
    Column('dbnode_id', Integer, ForeignKey('db_dbnode.id', deferrable=True, initially='DEFERRED')),
    Column('dbgroup_id', Integer, ForeignKey('db_dbgroup.id', deferrable=True, initially='DEFERRED')),
    UniqueConstraint('dbgroup_id', 'dbnode_id', name='db_dbgroup_dbnodes_dbgroup_id_dbnode_id_key'),
)


class DbGroup(Base):
    """Class to store groups using SQLA backend."""

    __tablename__ = 'db_dbgroup'

    id = Column(Integer, primary_key=True)  # pylint: disable=invalid-name

    uuid = Column(UUID(as_uuid=True), default=get_new_uuid, unique=True)
    label = Column(String(255), index=True)

    type_string = Column(String(255), default='', index=True)

    time = Column(DateTime(timezone=True), default=timezone.now)
    description = Column(Text, nullable=True)

    extras = Column(JSONB, default=dict, nullable=False)

    user_id = Column(Integer, ForeignKey('db_dbuser.id', ondelete='CASCADE', deferrable=True, initially='DEFERRED'))
    user = relationship('DbUser', backref=backref('dbgroups', cascade='merge'))

    dbnodes = relationship('DbNode', secondary=table_groups_nodes, backref='dbgroups', lazy='dynamic')

    __table_args__ = (UniqueConstraint('label', 'type_string'),)

    Index('db_dbgroup_dbnodes_dbnode_id_idx', table_groups_nodes.c.dbnode_id)
    Index('db_dbgroup_dbnodes_dbgroup_id_idx', table_groups_nodes.c.dbgroup_id)

    @property
    def pk(self):
        return self.id

    def __str__(self):
        return f'<DbGroup [type: {self.type_string}] "{self.label}">'



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=global-statement
"""Module with implementation of the database backend using Django."""
from aiida.backends.utils import create_sqlalchemy_engine, create_scoped_session_factory

ENGINE = None
SESSION_FACTORY = None


def reset_session():
    """Reset the session which means setting the global engine and session factory instances to `None`."""
    global ENGINE
    global SESSION_FACTORY

    if ENGINE is not None:
        ENGINE.dispose()

    if SESSION_FACTORY is not None:
        SESSION_FACTORY.expunge_all()  # pylint: disable=no-member
        SESSION_FACTORY.close()  # pylint: disable=no-member

    ENGINE = None
    SESSION_FACTORY = None


def get_scoped_session(**kwargs):
    """Return a scoped session for the given profile that is exclusively to be used for the `QueryBuilder`.

    Since the `QueryBuilder` implementation uses SqlAlchemy to map the query onto the models in order to generate the
    SQL to be sent to the database, it requires a session, which is an :class:`sqlalchemy.orm.session.Session` instance.
    The only purpose is for SqlAlchemy to be able to connect to the database perform the query and retrieve the results.
    Even the Django backend implementation will use SqlAlchemy for its `QueryBuilder` and so also needs an SqlA session.
    It is important that we do not reuse the scoped session factory in the SqlAlchemy implementation, because that runs
    the risk of cross-talk once profiles can be switched dynamically in a single python interpreter. Therefore the
    Django implementation of the `QueryBuilder` should keep its own SqlAlchemy engine and scoped session factory
    instances that are used to provide the query builder with a session.

    :param kwargs: keyword arguments that will be passed on to :py:func:`aiida.backends.utils.create_sqlalchemy_engine`,
        opening the possibility to change QueuePool time outs and more.
        See https://docs.sqlalchemy.org/en/13/core/engines.html?highlight=create_engine#sqlalchemy.create_engine for
        more info.

    :return: :class:`sqlalchemy.orm.session.Session` instance with engine configured for the given profile.
    """
    from aiida.manage.configuration import get_profile

    global ENGINE
    global SESSION_FACTORY

    if SESSION_FACTORY is not None:
        session = SESSION_FACTORY()
        return session

    if ENGINE is None:
        ENGINE = create_sqlalchemy_engine(get_profile(), **kwargs)

    SESSION_FACTORY = create_scoped_session_factory(ENGINE)

    return SESSION_FACTORY()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module for implementations of database backends."""

BACKEND_DJANGO = 'django'
BACKEND_SQLA = 'sqlalchemy'


def get_backend_manager(backend):
    """Get an instance of the `BackendManager` for the current backend.

    :param backend: the type of the database backend
    :return: `BackendManager`
    """
    if backend == BACKEND_DJANGO:
        from aiida.backends.djsite.manager import DjangoBackendManager
        return DjangoBackendManager()

    if backend == BACKEND_SQLA:
        from aiida.backends.sqlalchemy.manager import SqlaBackendManager
        return SqlaBackendManager()

    raise Exception(f'unknown backend type `{backend}`')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Classes for describing atomic orbitals.

Contains general Orbital class.
For subclasses of Orbital, see submodules.
"""

from aiida.common.exceptions import ValidationError
from aiida.plugins.entry_point import get_entry_point_from_class


def validate_int(value):
    """
    Validate that the value is an int
    """
    try:
        conv_value = int(value)
    except ValueError:
        raise ValidationError('must be an int number')
    return conv_value


def validate_int_or_none(value):
    """
    Validate that the value is a int or is None
    """
    if value is None:
        return None
    return validate_int(value)


def validate_float(value):
    """
    Validate that the value is a float
    """
    try:
        conv_value = float(value)
    except ValueError:
        raise ValidationError('must be a float number')
    return conv_value


def validate_float_or_none(value):
    """
    Validate that the value is a float or is None
    """
    if value is None:
        return None
    return validate_float(value)


def validate_len3_list(value):
    """
    Validate that the value is a list of three floats
    """
    try:
        conv_value = list(float(i) for i in value)
        if len(conv_value) != 3:
            raise ValueError
    except (ValueError, TypeError):
        raise ValidationError('must be a list of three float number')
    return conv_value


def validate_len3_list_or_none(value):
    """
    Validate that the value is a list of three floats or is None
    """
    if value is None:
        return None
    return validate_len3_list(value)


class Orbital:
    """
    Base class for Orbitals. Can handle certain basic fields, their setting
    and validation. More complex Orbital objects should then inherit from
    this class

    :param position: the absolute position (three floats) units in angstrom
    :param x_orientation: x,y,z unit vector defining polar angle theta
                          in spherical coordinates unitless
    :param z_orientation: x,y,z unit vector defining azimuthal angle phi
                          in spherical coordinates unitless
    :param orientation_spin: x,y,z unit vector defining the spin orientation
                             unitless
    :param diffusivity: Float controls the radial term in orbital equation
                        units are reciprocal Angstrom.
    """
    # len-2 tuples, with name and validator function
    _base_fields_required = (('position', validate_len3_list),
                             #NOTE: _orbital_type is internally used to manage the orbital type
                             )

    # len-3 tuples, with (name, validator, default_value)
    # See how it is defined in the RealhydrogenOrbital class
    _base_fields_optional = tuple()

    def __init__(self, **kwargs):
        # This runs the validator as well
        self.set_orbital_dict(kwargs)

    def __repr__(self):
        return f'<{self.__class__.__name__}: {str(self)}>'

    def _validate_keys(self, input_dict):
        """
        Checks all the input_dict and tries to validate them, to ensure
        that they have been properly set raises Exceptions indicating any
        problems that should arise during the validation

        :param input_dict: a dictionary of inputs
        :return: input_dict: the original dictionary with all validated kyes
                 now removed
        :return: validated_dict: a dictionary containing all the input keys
                 which have now been validated.
        """

        validated_dict = {}
        if '_orbital_type' in input_dict:
            raise ValidationError('You cannot manually set the _orbital_type')
        entry_point = get_entry_point_from_class(self.__class__.__module__, self.__class__.__name__)[1]
        if entry_point is None:
            raise ValidationError(
                'Unable to detect entry point for current class {}, maybe you did not register an entry point for it?'.
                format(self.__class__)
            )

        validated_dict['_orbital_type'] = entry_point.name

        for name, validator in self._base_fields_required:
            try:
                value = input_dict.pop(name)
            except KeyError:
                raise ValidationError(f"Missing required parameter '{name}'")
            # This might raise ValidationError
            try:
                value = validator(value)
            except ValidationError as exc:
                raise exc.__class__(f"Error validating '{name}': {str(exc)}")
            validated_dict[name] = value

        for name, validator, default_value in self._base_fields_optional:
            try:
                value = input_dict.pop(name)
            except KeyError:
                value = default_value
            # This might raise ValidationError
            try:
                value = validator(value)
            except ValidationError as exc:
                raise exc.__class__(f"Error validating '{name}': {str(exc)}")
            validated_dict[name] = value

        if input_dict:
            raise ValidationError(f'Unknown keys: {list(input_dict.keys())}')
        return validated_dict

    def set_orbital_dict(self, init_dict):
        """
        Sets the orbital_dict, which can vary depending on the particular
        implementation of this base class.

        :param init_dict: the initialization dictionary
        """
        self._orbital_dict = self._validate_keys(init_dict)

    def get_orbital_dict(self):
        """
        returns the internal keys as a dictionary
        """
        return self._orbital_dict



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi data singlefile` command."""

from aiida.cmdline.commands.cmd_data import verdi_data
from aiida.cmdline.params import arguments, types
from aiida.cmdline.utils import decorators, echo


@verdi_data.group('singlefile')
def singlefile():
    """Work with SinglefileData nodes."""


@singlefile.command('content')
@arguments.DATUM(type=types.DataParamType(sub_classes=('aiida.data:singlefile',)))
@decorators.with_dbenv()
def singlefile_content(datum):
    """Show the content of the file."""
    try:
        echo.echo(datum.get_content())
    except (IOError, OSError) as exception:
        echo.echo_critical(f'could not read the content for SinglefileData<{datum.pk}>: {str(exception)}')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Reusable command line interface options for the setup commands."""
import functools
import getpass
import hashlib

import click

from aiida.backends import BACKEND_DJANGO
from aiida.cmdline.params import options, types
from aiida.manage.configuration import get_config, get_config_option, Profile
from aiida.manage.external.postgres import DEFAULT_DBINFO
from aiida.manage.external.rmq import BROKER_DEFAULTS

PASSWORD_UNCHANGED = '***'  # noqa


def validate_profile_parameter(ctx):
    """Validate that the context contains the option `profile` and it contains a `Profile` instance.

    :param ctx: click context which should contain the selected profile
    :raises: BadParameter if the context does not contain a `Profile` instance for option `profile`
    """
    option = 'profile'
    if option not in ctx.params or ctx.params[option] is None or not isinstance(ctx.params[option], Profile):
        raise click.BadParameter('specifying the name of the profile is required', param_hint=f'"--{option}"')


def get_profile_attribute_default(attribute_tuple, ctx):
    """Return the default value for the given attribute of the profile passed in the context.

    :param attribute: attribute for which to get the current value
    :param ctx: click context which should contain the selected profile
    :return: profile attribute default value if set, or None
    """
    attribute, default = attribute_tuple

    try:
        validate_profile_parameter(ctx)
    except click.BadParameter:
        return default
    else:
        try:
            return getattr(ctx.params['profile'], attribute)
        except KeyError:
            return default


def get_repository_uri_default(ctx):
    """Return the default value for the repository URI for the current profile in the click context.

    :param ctx: click context which should contain the selected profile
    :return: default repository URI
    """
    import os
    from aiida.manage.configuration.settings import AIIDA_CONFIG_FOLDER

    validate_profile_parameter(ctx)

    return os.path.join(AIIDA_CONFIG_FOLDER, 'repository', ctx.params['profile'].name)


def get_quicksetup_repository_uri(ctx, param, value):  # pylint: disable=unused-argument
    """Return the repository URI to be used as default in `verdi quicksetup`

    :param ctx: click context which should contain the contextual parameters
    :return: the repository URI
    """
    return get_repository_uri_default(ctx)


def get_quicksetup_database_name(ctx, param, value):  # pylint: disable=unused-argument
    """Determine the database name to be used as default for the Postgres connection in `verdi quicksetup`

    If a value is explicitly passed, that value is returned unchanged.

    If no value is passed, the name will be <profile_name>_<os_user>_<hash>, where <os_user> is the name of the current
    operating system user and <hash> is a hash of the path of the configuration directory.

    Note: This ensures that profiles named ``test_...`` will have databases named ``test_...`` .

    :param ctx: click context which should contain the contextual parameters
    :return: the database name
    """
    if value is not None:
        return value

    config = get_config()
    profile = ctx.params['profile'].name
    config_hash = hashlib.md5(config.dirpath.encode('utf-8')).hexdigest()
    database_name = f'{profile}_{getpass.getuser()}_{config_hash}'

    return database_name


def get_quicksetup_username(ctx, param, value):  # pylint: disable=unused-argument
    """Determine the username to be used as default for the Postgres connection in `verdi quicksetup`

    If a value is explicitly passed, that value is returned. If there is no value, the name will be based on the
    name of the current operating system user and the hash of the path of the configuration directory.

    :param ctx: click context which should contain the contextual parameters
    :return: the username
    """
    if value is not None:
        return value

    config = get_config()
    config_hash = hashlib.md5(config.dirpath.encode('utf-8')).hexdigest()
    username = f'aiida_qs_{getpass.getuser()}_{config_hash}'

    return username


def get_quicksetup_password(ctx, param, value):  # pylint: disable=unused-argument
    """Determine the password to be used as default for the Postgres connection in `verdi quicksetup`

    If a value is explicitly passed, that value is returned. If there is no value, the current username in the context
    will be scanned for in currently existing profiles. If it does, the corresponding password will be used. If no such
    user already exists, a random password will be generated.

    :param ctx: click context which should contain the contextual parameters
    :return: the password
    """
    from aiida.common.hashing import get_random_string

    if value is not None:
        return value

    username = ctx.params['db_username']
    config = get_config()

    for available_profile in config.profiles:
        if available_profile.database_username == username:
            value = available_profile.database_password
            break
    else:
        value = get_random_string(16)

    return value


SETUP_PROFILE = options.OverridableOption(
    '--profile',
    prompt='Profile name',
    help='The name of the new profile.',
    required=True,
    type=types.ProfileParamType(cannot_exist=True),
    cls=options.interactive.InteractiveOption
)

SETUP_USER_EMAIL = options.USER_EMAIL.clone(
    prompt='Email Address (for sharing data)',
    default=get_config_option('user.email'),
    required_fn=lambda x: get_config_option('user.email') is None,
    required=True,
    cls=options.interactive.InteractiveOption
)

SETUP_USER_FIRST_NAME = options.USER_FIRST_NAME.clone(
    prompt='First name',
    default=get_config_option('user.first_name'),
    required_fn=lambda x: get_config_option('user.first_name') is None,
    required=True,
    cls=options.interactive.InteractiveOption
)

SETUP_USER_LAST_NAME = options.USER_LAST_NAME.clone(
    prompt='Last name',
    default=get_config_option('user.last_name'),
    required_fn=lambda x: get_config_option('user.last_name') is None,
    required=True,
    cls=options.interactive.InteractiveOption
)

SETUP_USER_INSTITUTION = options.USER_INSTITUTION.clone(
    prompt='Institution',
    default=get_config_option('user.institution'),
    required_fn=lambda x: get_config_option('user.institution') is None,
    required=True,
    cls=options.interactive.InteractiveOption
)

QUICKSETUP_DATABASE_ENGINE = options.DB_ENGINE

QUICKSETUP_DATABASE_BACKEND = options.DB_BACKEND

QUICKSETUP_DATABASE_HOSTNAME = options.DB_HOST

QUICKSETUP_DATABASE_PORT = options.DB_PORT

QUICKSETUP_DATABASE_NAME = options.OverridableOption(
    '--db-name',
    help='Name of the database to create.',
    type=types.NonEmptyStringParamType(),
    callback=get_quicksetup_database_name
)

QUICKSETUP_DATABASE_USERNAME = options.DB_USERNAME.clone(
    help='Name of the database user to create.', callback=get_quicksetup_username
)

QUICKSETUP_DATABASE_PASSWORD = options.DB_PASSWORD.clone(callback=get_quicksetup_password)

QUICKSETUP_SUPERUSER_DATABASE_USERNAME = options.OverridableOption(
    '--su-db-username', help='User name of the database super user.', type=click.STRING, default=DEFAULT_DBINFO['user']
)

QUICKSETUP_SUPERUSER_DATABASE_NAME = options.OverridableOption(
    '--su-db-name',
    help='Name of the template database to connect to as the database superuser.',
    type=click.STRING,
    default=DEFAULT_DBINFO['database']
)

QUICKSETUP_SUPERUSER_DATABASE_PASSWORD = options.OverridableOption(
    '--su-db-password',
    help='Password to connect as the database superuser.',
    type=click.STRING,
    hide_input=True,
    default=DEFAULT_DBINFO['password'],
)

QUICKSETUP_BROKER_PROTOCOL = options.BROKER_PROTOCOL

QUICKSETUP_BROKER_USERNAME = options.BROKER_USERNAME

QUICKSETUP_BROKER_PASSWORD = options.BROKER_PASSWORD

QUICKSETUP_BROKER_HOST = options.BROKER_HOST

QUICKSETUP_BROKER_PORT = options.BROKER_PORT

QUICKSETUP_BROKER_VIRTUAL_HOST = options.BROKER_VIRTUAL_HOST

QUICKSETUP_REPOSITORY_URI = options.REPOSITORY_PATH.clone(
    callback=get_quicksetup_repository_uri  # Cannot use `default` because `ctx` is needed to determine the default
)

SETUP_DATABASE_ENGINE = QUICKSETUP_DATABASE_ENGINE.clone(
    prompt='Database engine',
    contextual_default=functools.partial(get_profile_attribute_default, ('database_engine', 'postgresql_psycopg2')),
    cls=options.interactive.InteractiveOption
)

SETUP_DATABASE_BACKEND = QUICKSETUP_DATABASE_BACKEND.clone(
    prompt='Database backend',
    contextual_default=functools.partial(get_profile_attribute_default, ('database_backend', BACKEND_DJANGO)),
    cls=options.interactive.InteractiveOption
)

SETUP_DATABASE_HOSTNAME = QUICKSETUP_DATABASE_HOSTNAME.clone(
    prompt='Database host',
    contextual_default=functools.partial(get_profile_attribute_default, ('database_hostname', DEFAULT_DBINFO['host'])),
    cls=options.interactive.InteractiveOption
)

SETUP_DATABASE_PORT = QUICKSETUP_DATABASE_PORT.clone(
    prompt='Database port',
    contextual_default=functools.partial(get_profile_attribute_default, ('database_port', DEFAULT_DBINFO['port'])),
    cls=options.interactive.InteractiveOption
)

SETUP_DATABASE_NAME = QUICKSETUP_DATABASE_NAME.clone(
    prompt='Database name',
    required=True,
    contextual_default=functools.partial(get_profile_attribute_default, ('database_name', None)),
    cls=options.interactive.InteractiveOption
)

SETUP_DATABASE_USERNAME = QUICKSETUP_DATABASE_USERNAME.clone(
    prompt='Database username',
    required=True,
    contextual_default=functools.partial(get_profile_attribute_default, ('database_username', None)),
    cls=options.interactive.InteractiveOption
)

SETUP_DATABASE_PASSWORD = QUICKSETUP_DATABASE_PASSWORD.clone(
    prompt='Database password',
    required=True,
    contextual_default=functools.partial(get_profile_attribute_default, ('database_password', None)),
    cls=options.interactive.InteractiveOption
)

SETUP_BROKER_PROTOCOL = QUICKSETUP_BROKER_PROTOCOL.clone(
    prompt='Broker protocol',
    required=True,
    contextual_default=functools.partial(get_profile_attribute_default, ('broker_protocol', BROKER_DEFAULTS.protocol)),
    cls=options.interactive.InteractiveOption
)

SETUP_BROKER_USERNAME = QUICKSETUP_BROKER_USERNAME.clone(
    prompt='Broker username',
    required=True,
    contextual_default=functools.partial(get_profile_attribute_default, ('broker_username', BROKER_DEFAULTS.username)),
    cls=options.interactive.InteractiveOption
)

SETUP_BROKER_PASSWORD = QUICKSETUP_BROKER_PASSWORD.clone(
    prompt='Broker password',
    required=True,
    contextual_default=functools.partial(get_profile_attribute_default, ('broker_password', BROKER_DEFAULTS.password)),
    cls=options.interactive.InteractiveOption
)

SETUP_BROKER_HOST = QUICKSETUP_BROKER_HOST.clone(
    prompt='Broker host',
    required=True,
    contextual_default=functools.partial(get_profile_attribute_default, ('broker_host', BROKER_DEFAULTS.host)),
    cls=options.interactive.InteractiveOption
)

SETUP_BROKER_PORT = QUICKSETUP_BROKER_PORT.clone(
    prompt='Broker port',
    required=True,
    contextual_default=functools.partial(get_profile_attribute_default, ('broker_port', BROKER_DEFAULTS.port)),
    cls=options.interactive.InteractiveOption
)

SETUP_BROKER_VIRTUAL_HOST = QUICKSETUP_BROKER_VIRTUAL_HOST.clone(
    prompt='Broker virtual host name',
    required=True,
    contextual_default=functools.partial(
        get_profile_attribute_default, ('broker_virtual_host', BROKER_DEFAULTS.virtual_host)
    ),
    cls=options.interactive.InteractiveOption
)

SETUP_REPOSITORY_URI = QUICKSETUP_REPOSITORY_URI.clone(
    prompt='Repository directory',
    callback=None,  # Unset the `callback` to define the default, which is instead done by the `contextual_default`
    contextual_default=get_repository_uri_default,
    cls=options.interactive.InteractiveOption
)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Miscellaneous generic utility functions and classes."""
import filecmp
import inspect
import io
import os
import re
import sys
from uuid import UUID

from .lang import classproperty


def get_new_uuid():
    """
    Return a new UUID (typically to be used for new nodes).
    It uses the UUID version specified in
    aiida.backends.settings.AIIDANODES_UUID_VERSION
    """
    import uuid
    return str(uuid.uuid4())


def validate_uuid(given_uuid: str) -> bool:
    """A simple check for the UUID validity."""
    try:
        parsed_uuid = UUID(given_uuid, version=4)
    except ValueError:
        # If not a valid UUID
        return False

    # Check if there was any kind of conversion of the hex during
    # the validation
    return str(parsed_uuid) == given_uuid


def validate_list_of_string_tuples(val, tuple_length):
    """
    Check that:

    1. ``val`` is a list or tuple
    2. each element of the list:

      a. is a list or tuple
      b. is of length equal to the parameter tuple_length
      c. each of the two elements is a string

    Return if valid, raise ValidationError if invalid
    """
    from aiida.common.exceptions import ValidationError

    err_msg = (
        'the value must be a list (or tuple) '
        'of length-N list (or tuples), whose elements are strings; '
        'N={}'.format(tuple_length)
    )

    if not isinstance(val, (list, tuple)):
        raise ValidationError(err_msg)

    for element in val:
        if (
            not isinstance(element, (list, tuple)) or (len(element) != tuple_length) or
            not all(isinstance(s, str) for s in element)
        ):
            raise ValidationError(err_msg)

    return True


def get_unique_filename(filename, list_of_filenames):
    """
    Return a unique filename that can be added to the list_of_filenames.

    If filename is not in list_of_filenames, it simply returns the filename
    string itself. Otherwise, it appends a integer number to the filename
    (before the extension) until it finds a unique filename.

    :param filename: the filename to add
    :param list_of_filenames: the list of filenames to which filename
        should be added, without name duplicates

    :returns: Either filename or its modification, with a number appended
        between the name and the extension.
    """
    if filename not in list_of_filenames:
        return filename

    basename, ext = os.path.splitext(filename)

    # Not optimized, but for the moment this should be fast enough
    append_int = 1
    while True:
        new_filename = f'{basename:s}-{append_int:d}{ext:s}'
        if new_filename not in list_of_filenames:
            break
        append_int += 1
    return new_filename


def str_timedelta(dt, max_num_fields=3, short=False, negative_to_zero=False):  # pylint: disable=invalid-name
    """
    Given a dt in seconds, return it in a HH:MM:SS format.

    :param dt: a TimeDelta object
    :param max_num_fields: maximum number of non-zero fields to show
        (for instance if the number of days is non-zero, shows only
        days, hours and minutes, but not seconds)
    :param short: if False, print always ``max_num_fields`` fields, even
        if they are zero. If True, do not print the first fields, if they
        are zero.
    :param negative_to_zero: if True, set dt = 0 if dt < 0.
    """
    if max_num_fields <= 0:
        raise ValueError('max_num_fields must be > 0')

    s_tot = dt.total_seconds()  # Important to get more than 1 day, and for
    # negative values. dt.seconds would give
    # wrong results in these cases, see
    # http://docs.python.org/2/library/datetime.html
    s_tot = int(s_tot)

    if negative_to_zero:
        if s_tot < 0:
            s_tot = 0

    negative = (s_tot < 0)
    s_tot = abs(s_tot)

    negative_string = ' in the future' if negative else ' ago'

    # For the moment stay away from months and years, difficult to get
    days, remainder = divmod(s_tot, 3600 * 24)
    hours, remainder = divmod(remainder, 3600)
    minutes, seconds = divmod(remainder, 60)

    all_fields = [(days, 'D'), (hours, 'h'), (minutes, 'm'), (seconds, 's')]
    fields = []
    start_insert = False
    counter = 0
    for idx, field in enumerate(all_fields):
        if field[0] != 0:
            start_insert = True
        if (len(all_fields) - idx) <= max_num_fields:
            start_insert = True
        if start_insert:
            if counter >= max_num_fields:
                break
            fields.append(field)
            counter += 1

    if short:
        while len(fields) > 1:  # at least one element has to remain
            if fields[0][0] != 0:
                break
            fields.pop(0)  # remove first element

    # Join the fields
    raw_string = ':'.join(['{:02d}{}'.format(*f) for f in fields])

    if raw_string.startswith('0'):
        raw_string = raw_string[1:]

    # Return the resulting string, appending a suitable string if the time
    # is negative
    return f'{raw_string}{negative_string}'


def get_class_string(obj):
    """
    Return the string identifying the class of the object (module + object name,
    joined by dots).

    It works both for classes and for class instances.
    """
    if inspect.isclass(obj):
        return f'{obj.__module__}.{obj.__name__}'

    return f'{obj.__module__}.{obj.__class__.__name__}'


def get_object_from_string(class_string):
    """
    Given a string identifying an object (as returned by the get_class_string
    method) load and return the actual object.
    """
    import importlib

    the_module, _, the_name = class_string.rpartition('.')

    return getattr(importlib.import_module(the_module), the_name)


def grouper(n, iterable):  # pylint: disable=invalid-name
    """
    Given an iterable, returns an iterable that returns tuples of groups of
    elements from iterable of length n, except the last one that has the
    required length to exaust iterable (i.e., there is no filling applied).

    :param n: length of each tuple (except the last one,that will have length
       <= n
    :param iterable: the iterable to divide in groups
    """
    import itertools

    iterator = iter(iterable)
    while True:
        chunk = tuple(itertools.islice(iterator, n))
        if not chunk:
            return
        yield chunk


class ArrayCounter:
    """
    A counter & a method that increments it and returns its value.
    It is used in various tests.
    """
    seq = None

    def __init__(self):
        self.seq = -1

    def array_counter(self):
        self.seq += 1
        return self.seq


def are_dir_trees_equal(dir1, dir2):
    """
    Compare two directories recursively. Files in each directory are
    assumed to be equal if their names and contents are equal.

    @param dir1: First directory path
    @param dir2: Second directory path

    @return: True if the directory trees are the same and
        there were no errors while accessing the directories or files,
        False otherwise.
    """

    # Directory comparison
    dirs_cmp = filecmp.dircmp(dir1, dir2)
    if dirs_cmp.left_only or dirs_cmp.right_only or dirs_cmp.funny_files:
        return (
            False, 'Left directory: {}, right directory: {}, files only '
            'in left directory: {}, files only in right directory: '
            '{}, not comparable files: {}'.format(
                dir1, dir2, dirs_cmp.left_only, dirs_cmp.right_only, dirs_cmp.funny_files
            )
        )

    # If the directories contain the same files, compare the common files
    (_, mismatch, errors) = filecmp.cmpfiles(dir1, dir2, dirs_cmp.common_files, shallow=False)
    if mismatch:
        return (
            False, 'The following files in the directories {} and {} '
            "don't match: {}".format(dir1, dir2, mismatch)
        )
    if errors:
        return (
            False, 'The following files in the directories {} and {} '
            "aren't regular: {}".format(dir1, dir2, errors)
        )

    for common_dir in dirs_cmp.common_dirs:
        new_dir1 = os.path.join(dir1, common_dir)
        new_dir2 = os.path.join(dir2, common_dir)
        res, msg = are_dir_trees_equal(new_dir1, new_dir2)
        if not res:
            return False, msg

    return True, f'The given directories ({dir1} and {dir2}) are equal'


class Prettifier:
    """
    Class to manage prettifiers (typically for labels of kpoints
    in band plots)
    """

    @classmethod
    def _prettify_label_pass(cls, label):
        """
        No-op prettifier, simply returns  the same label

        :param label: a string to prettify
        """
        return label

    @classmethod
    def _prettify_label_agr(cls, label):
        """
        Prettifier for XMGrace

        :param label: a string to prettify
        """

        label = (
            label
                .replace('GAMMA', r'\xG\f{}')
                .replace('DELTA', r'\xD\f{}')
                .replace('LAMBDA', r'\xL\f{}')
                .replace('SIGMA', r'\xS\f{}')
        )  # yapf:disable
        return re.sub(r'_(.?)', r'\\s\1\\N', label)

    @classmethod
    def _prettify_label_agr_simple(cls, label):
        """
        Prettifier for XMGrace (for old label names)

        :param label: a string to prettify
        """

        if label == 'G':
            return r'\xG'

        return re.sub(r'(\d+)', r'\\s\1\\N', label)

    @classmethod
    def _prettify_label_gnuplot(cls, label):
        """
        Prettifier for Gnuplot

        :note: uses unicode, returns unicode strings (potentially, if needed)

        :param label: a string to prettify
        """

        label = (
            label
                .replace('GAMMA', 'Γ')
                .replace('DELTA', 'Δ')
                .replace('LAMBDA', 'Λ')
                .replace('SIGMA', 'Σ')
        )  # yapf:disable
        return re.sub(r'_(.?)', r'_{\1}', label)

    @classmethod
    def _prettify_label_gnuplot_simple(cls, label):
        """
        Prettifier for Gnuplot (for old label names)

        :note: uses unicode, returns unicode strings (potentially, if needed)

        :param label: a string to prettify
        """

        if label == 'G':
            return 'Γ'

        return re.sub(r'(\d+)', r'_{\1}', label)

    @classmethod
    def _prettify_label_latex(cls, label):
        """
        Prettifier for matplotlib, using LaTeX syntax

        :param label: a string to prettify
        """

        label = (
            label
                .replace('GAMMA', r'$\Gamma$')
                .replace('DELTA', r'$\Delta$')
                .replace('LAMBDA', r'$\Lambda$')
                .replace('SIGMA', r'$\Sigma$')
        )  # yapf:disable
        label = re.sub(r'_(.?)', r'$_{\1}$', label)

        # label += r"$_{\vphantom{0}}$"

        return label

    @classmethod
    def _prettify_label_latex_simple(cls, label):
        """
        Prettifier for matplotlib, using LaTeX syntax (for old label names)

        :param label: a string to prettify
        """
        if label == 'G':
            return r'$\Gamma$'

        return re.sub(r'(\d+)', r'$_{\1}$', label)

    @classproperty
    def prettifiers(cls):  # pylint: disable=no-self-argument
        """
        Property that returns a dictionary that for each string associates
        the function to prettify a label

        :return: a dictionary where keys are strings and values are functions
        """
        return {
            'agr_seekpath': cls._prettify_label_agr,
            'agr_simple': cls._prettify_label_agr_simple,
            'latex_simple': cls._prettify_label_latex_simple,
            'latex_seekpath': cls._prettify_label_latex,
            'gnuplot_simple': cls._prettify_label_gnuplot_simple,
            'gnuplot_seekpath': cls._prettify_label_gnuplot,
            'pass': cls._prettify_label_pass,
        }

    @classmethod
    def get_prettifiers(cls):
        """
        Return a list of valid prettifier strings

        :return: a list of strings
        """
        return sorted(cls.prettifiers.keys())  # pylint: disable=no-member

    def __init__(self, format):  # pylint: disable=redefined-builtin
        """
        Create a class to pretttify strings of a given format

        :param format: a string with the format to use to prettify.
           Valid formats are obtained from self.prettifiers
        """
        if format is None:
            format = 'pass'

        try:
            self._prettifier_f = self.prettifiers[format]  # pylint: disable=unsubscriptable-object
        except KeyError:
            raise ValueError(f"Unknown prettifier format {format}; valid formats: {', '.join(self.get_prettifiers())}")

    def prettify(self, label):
        """
        Prettify a label using the format passed in the initializer

        :param label: the string to prettify
        :return: a prettified string
        """
        return self._prettifier_f(label)


def prettify_labels(labels, format=None):  # pylint: disable=redefined-builtin
    """
    Prettify label for typesetting in various formats

    :param labels: a list of length-2 tuples, in the format(position, label)
    :param format: a string with the format for the prettifier (e.g. 'agr',
         'matplotlib', ...)
    :return: the same list as labels, but with the second value possibly replaced
         with a prettified version that typesets nicely in the selected format
    """
    prettifier = Prettifier(format)

    return [(pos, prettifier.prettify(label)) for pos, label in labels]


def join_labels(labels, join_symbol='|', threshold=1.e-6):
    """
    Join labels with a joining symbol when they are very close

    :param labels: a list of length-2 tuples, in the format(position, label)
    :param join_symbol: the string to use to join different paths. By default, a pipe
    :param threshold: the threshold to decide if two float values are the same and should
         be joined
    :return: the same list as labels, but with the second value possibly replaced
         with strings joined when close enough
    """
    if labels:
        new_labels = [list(labels[0])]
        # modify labels when in overlapping position
        j = 0
        for i in range(1, len(labels)):
            if abs(labels[i][0] - labels[i - 1][0]) < threshold:
                new_labels[j][1] += join_symbol + labels[i][1]
            else:
                new_labels.append(list(labels[i]))
                j += 1
    else:
        new_labels = []

    return new_labels


def strip_prefix(full_string, prefix):
    """
    Strip the prefix from the given string and return it. If the prefix is not present
    the original string will be returned unaltered

    :param full_string: the string from which to remove the prefix
    :param prefix: the prefix to remove
    :return: the string with prefix removed
    """
    if full_string.startswith(prefix):
        return full_string.rsplit(prefix)[1]

    return full_string


class Capturing:
    """
    This class captures stdout and returns it
    (as a list, split by lines).

    Note: if you raise a SystemExit, you have to catch it outside.
    E.g., in our tests, this works::

        import sys
        with self.assertRaises(SystemExit):
            with Capturing() as output:
                sys.exit()

    But out of the testing environment, the code instead just exits.

    To use it, access the obj.stdout_lines, or just iterate over the object

    :param capture_stderr: if True, also captures sys.stderr. To access the
        lines, use obj.stderr_lines. If False, obj.stderr_lines is None.
    """

    # pylint: disable=attribute-defined-outside-init

    def __init__(self, capture_stderr=False):
        self.stdout_lines = list()
        super().__init__()

        self._capture_stderr = capture_stderr
        if self._capture_stderr:
            self.stderr_lines = list()
        else:
            self.stderr_lines = None

    def __enter__(self):
        """Enter the context where all output is captured."""
        self._stdout = sys.stdout
        self._stringioout = io.StringIO()
        sys.stdout = self._stringioout
        if self._capture_stderr:
            self._stderr = sys.stderr
            self._stringioerr = io.StringIO()
            sys.stderr = self._stringioerr
        return self

    def __exit__(self, *args):
        """Exit the context where all output is captured."""
        self.stdout_lines.extend(self._stringioout.getvalue().splitlines())
        sys.stdout = self._stdout
        del self._stringioout  # free up some memory
        if self._capture_stderr:
            self.stderr_lines.extend(self._stringioerr.getvalue().splitlines())
            sys.stderr = self._stderr
            del self._stringioerr  # free up some memory

    def __str__(self):
        return str(self.stdout_lines)

    def __iter__(self):
        return iter(self.stdout_lines)


class ErrorAccumulator:
    """
    Allows to run a number of functions and collect all the errors they raise

    This allows to validate multiple things and tell the user about all the
    errors encountered at once. Works best if the individual functions do not depend on each other.

    Does not allow to trace the stack of each error, therefore do not use for debugging, but for
    semantical checking with user friendly error messages.
    """

    def __init__(self, *error_cls):
        self.error_cls = error_cls
        self.errors = {k: [] for k in self.error_cls}

    def run(self, function, *args, **kwargs):
        try:
            function(*args, **kwargs)
        except self.error_cls as err:
            self.errors[err.__class__].append(err)

    def success(self):
        return bool(not any(self.errors.values()))

    def result(self, raise_error=Exception):
        if raise_error:
            self.raise_errors(raise_error)
        return self.success(), self.errors

    def raise_errors(self, raise_cls):
        if not self.success():
            raise raise_cls(f'The following errors were encountered: {self.errors}')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Manage AiiDA queries."""
import abc


class AbstractQueryManager(abc.ABC):
    """Manage AiiDA queries."""

    def __init__(self, backend):
        """
        :param backend: The AiiDA backend
        :type backend: :class:`aiida.orm.implementation.sql.backends.SqlBackend`
        """
        self._backend = backend

    def get_duplicate_uuids(self, table):
        """
        Return a list of rows with identical UUID

        :param table: Database table with uuid column, e.g. 'db_dbnode'
        :type str:

        :return: list of tuples of (id, uuid) of rows with duplicate UUIDs
        :rtype list:
        """
        query = f"""
            SELECT s.id, s.uuid FROM (SELECT *, COUNT(*) OVER(PARTITION BY uuid) AS c FROM {table})
            AS s WHERE c > 1
            """
        return self._backend.execute_raw(query)

    def apply_new_uuid_mapping(self, table, mapping):
        for pk, uuid in mapping.items():
            query = f"""UPDATE {table} SET uuid = '{uuid}' WHERE id = {pk}"""
            with self._backend.cursor() as cursor:
                cursor.execute(query)

    @staticmethod
    def get_creation_statistics(user_pk=None):
        """
        Return a dictionary with the statistics of node creation, summarized by day.

        :note: Days when no nodes were created are not present in the returned `ctime_by_day` dictionary.

        :param user_pk: If None (default), return statistics for all users.
            If user pk is specified, return only the statistics for the given user.

        :return: a dictionary as
            follows::

                {
                   "total": TOTAL_NUM_OF_NODES,
                   "types": {TYPESTRING1: count, TYPESTRING2: count, ...},
                   "ctime_by_day": {'YYYY-MMM-DD': count, ...}

            where in `ctime_by_day` the key is a string in the format 'YYYY-MM-DD' and the value is
            an integer with the number of nodes created that day.
        """
        import datetime
        from collections import Counter
        from aiida.orm import User, Node, QueryBuilder

        def count_statistics(dataset):

            def get_statistics_dict(dataset):
                results = {}
                for count, typestring in sorted((v, k) for k, v in dataset.items())[::-1]:
                    results[typestring] = count
                return results

            count_dict = {}

            types = Counter([r[2] for r in dataset])
            count_dict['types'] = get_statistics_dict(types)

            ctimelist = [r[1].strftime('%Y-%m-%d') for r in dataset]
            ctime = Counter(ctimelist)

            if ctimelist:

                # For the way the string is formatted, we can just sort it alphabetically
                firstdate = datetime.datetime.strptime(sorted(ctimelist)[0], '%Y-%m-%d')
                lastdate = datetime.datetime.strptime(sorted(ctimelist)[-1], '%Y-%m-%d')

                curdate = firstdate
                outdata = {}

                while curdate <= lastdate:
                    curdatestring = curdate.strftime('%Y-%m-%d')
                    outdata[curdatestring] = ctime.get(curdatestring, 0)
                    curdate += datetime.timedelta(days=1)
                count_dict['ctime_by_day'] = outdata

            else:
                count_dict['ctime_by_day'] = {}

            return count_dict

        statistics = {}

        q_build = QueryBuilder()
        q_build.append(Node, project=['id', 'ctime', 'type'], tag='node')

        if user_pk is not None:
            q_build.append(User, with_node='node', project='email', filters={'pk': user_pk})
        qb_res = q_build.all()

        # total count
        statistics['total'] = len(qb_res)
        statistics.update(count_statistics(qb_res))

        return statistics

    @staticmethod
    def _extract_formula(akinds, asites, args):
        """
        Extract formula from the structure object.

        :param akinds: list of kinds, e.g. [{'mass': 55.845, 'name': 'Fe', 'symbols': ['Fe'], 'weights': [1.0]},
                                            {'mass': 15.9994, 'name': 'O', 'symbols': ['O'], 'weights': [1.0]}]
        :param asites: list of structure sites e.g. [{'position': [0.0, 0.0, 0.0], 'kind_name': 'Fe'},
                                                        {'position': [2.0, 2.0, 2.0], 'kind_name': 'O'}]
        :param args: a namespace with parsed command line parameters, here only 'element' and 'element_only' are used
        :type args: dict

        :return: a string with formula if the formula is found
        """
        from aiida.orm.nodes.data.structure import (get_formula, get_symbols_string)

        if args.element is not None:
            all_symbols = [_['symbols'][0] for _ in akinds]
            if not any([s in args.element for s in all_symbols]):
                return None

        if args.element_only is not None:
            all_symbols = [_['symbols'][0] for _ in akinds]
            if not all([s in all_symbols for s in args.element_only]):
                return None

        # We want only the StructureData that have attributes
        if akinds is None or asites is None:
            return '<<UNKNOWN>>'

        symbol_dict = {}
        for k in akinds:
            symbols = k['symbols']
            weights = k['weights']
            symbol_dict[k['name']] = get_symbols_string(symbols, weights)

        try:
            symbol_list = []
            for site in asites:
                symbol_list.append(symbol_dict[site['kind_name']])
            formula = get_formula(symbol_list, mode=args.formulamode)
        # If for some reason there is no kind with the name
        # referenced by the site
        except KeyError:
            formula = '<<UNKNOWN>>'
        return formula

    def get_bands_and_parents_structure(self, args):
        """Search for bands and return bands and the closest structure that is a parent of the instance.
        This is the backend independent way, can be overriden for performance reason

        :returns:
            A list of sublists, each latter containing (in order):
                pk as string, formula as string, creation date, bandsdata-label
        """
        # pylint: disable=too-many-locals

        import datetime
        from aiida.common import timezone
        from aiida import orm

        q_build = orm.QueryBuilder()
        if args.all_users is False:
            q_build.append(orm.User, tag='creator', filters={'email': orm.User.objects.get_default().email})
        else:
            q_build.append(orm.User, tag='creator')

        group_filters = {}

        if args.group_name is not None:
            group_filters.update({'name': {'in': args.group_name}})
        if args.group_pk is not None:
            group_filters.update({'id': {'in': args.group_pk}})

        q_build.append(orm.Group, tag='group', filters=group_filters, with_user='creator')

        bdata_filters = {}
        if args.past_days is not None:
            bdata_filters.update({'ctime': {'>=': timezone.now() - datetime.timedelta(days=args.past_days)}})

        q_build.append(
            orm.BandsData, tag='bdata', with_group='group', filters=bdata_filters, project=['id', 'label', 'ctime']
        )
        bands_list_data = q_build.all()

        q_build.append(
            orm.StructureData,
            tag='sdata',
            with_descendants='bdata',
            # We don't care about the creator of StructureData
            project=['id', 'attributes.kinds', 'attributes.sites']
        )

        q_build.order_by({orm.StructureData: {'ctime': 'desc'}})

        structure_dict = dict()
        list_data = q_build.distinct().all()
        for bid, _, _, _, akinds, asites in list_data:
            structure_dict[bid] = (akinds, asites)

        entry_list = []
        already_visited_bdata = set()

        for [bid, blabel, bdate] in bands_list_data:

            # We process only one StructureData per BandsData.
            # We want to process the closest StructureData to
            # every BandsData.
            # We hope that the StructureData with the latest
            # creation time is the closest one.
            # This will be updated when the QueryBuilder supports
            # order_by by the distance of two nodes.
            if already_visited_bdata.__contains__(bid):
                continue
            already_visited_bdata.add(bid)
            strct = structure_dict.get(bid, None)

            if strct is not None:
                akinds, asites = strct
                formula = self._extract_formula(akinds, asites, args)
            else:
                if args.element is not None or args.element_only is not None:
                    formula = None
                else:
                    formula = '<<NOT FOUND>>'

            if formula is None:
                continue
            entry_list.append([str(bid), str(formula), bdate.strftime('%d %b %Y'), blabel])

        return entry_list

    @staticmethod
    def get_all_parents(node_pks, return_values=('id',)):
        """Get all the parents of given nodes

        :param node_pks: one node pk or an iterable of node pks
        :return: a list of aiida objects with all the parents of the nodes"""
        from aiida.orm import Node, QueryBuilder

        q_build = QueryBuilder()
        q_build.append(Node, tag='low_node', filters={'id': {'in': node_pks}})
        q_build.append(Node, with_descendants='low_node', project=return_values)
        return q_build.all()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Command for `verdi plugins`."""
import inspect

import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.utils import decorators, echo
from aiida.plugins.entry_point import ENTRY_POINT_GROUP_TO_MODULE_PATH_MAP


@verdi.group('plugin')
def verdi_plugin():
    """Inspect AiiDA plugins."""


@verdi_plugin.command('list')
@click.argument('entry_point_group', type=click.Choice(ENTRY_POINT_GROUP_TO_MODULE_PATH_MAP.keys()), required=False)
@click.argument('entry_point', type=click.STRING, required=False)
@decorators.with_dbenv()
def plugin_list(entry_point_group, entry_point):
    """Display a list of all available plugins."""
    from aiida.common import EntryPointError
    from aiida.cmdline.utils.common import print_process_info
    from aiida.engine import Process
    from aiida.plugins.entry_point import get_entry_point_names, load_entry_point

    if entry_point_group is None:
        echo.echo_info('Available entry point groups:')
        for group in sorted(ENTRY_POINT_GROUP_TO_MODULE_PATH_MAP.keys()):
            echo.echo(f'* {group}')

        echo.echo('')
        echo.echo_info('Pass one of the groups as an additional argument to show the registered plugins')
        return

    if entry_point:
        try:
            plugin = load_entry_point(entry_point_group, entry_point)
        except EntryPointError as exception:
            echo.echo_critical(str(exception))
        else:
            try:
                if (inspect.isclass(plugin) and issubclass(plugin, Process)) or plugin.is_process_function:
                    print_process_info(plugin)
                else:
                    echo.echo(str(plugin.get_description()))
            except AttributeError:
                echo.echo_error(f'No description available for {entry_point}')
    else:
        entry_points = get_entry_point_names(entry_point_group)
        if entry_points:
            echo.echo(f'Registered entry points for {entry_point_group}:')
            for registered_entry_point in entry_points:
                echo.echo(f'* {registered_entry_point}')

            echo.echo('')
            echo.echo_info('Pass the entry point as an argument to display detailed information')
        else:
            echo.echo_error(f'No plugins found for group {entry_point_group}')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Implementation of RESTful API for AiiDA based on flask and flask_restful.

Author: Snehal P. Waychal and Fernando Gargiulo @ Theos, EPFL
"""

from flask import Flask, jsonify
from flask_restful import Api
from werkzeug.exceptions import HTTPException


class App(Flask):
    """
    Basic Flask App customized for this REST Api purposes
    """

    def __init__(self, *args, **kwargs):

        # Decide whether or not to catch the internal server exceptions (
        # default is True)
        catch_internal_server = True
        try:
            catch_internal_server = kwargs.pop('catch_internal_server')
        except KeyError:
            pass

        # Basic initialization
        super().__init__(*args, **kwargs)

        # Error handler
        from aiida.restapi.common.exceptions import RestInputValidationError, \
            RestValidationError, RestFeatureNotAvailable

        if catch_internal_server:

            @self.errorhandler(Exception)
            def error_handler(error):
                # pylint: disable=unused-variable
                """Error handler to return customized error messages from rest api"""

                if isinstance(error, RestValidationError):
                    response = jsonify({'message': str(error)})
                    response.status_code = 400

                elif isinstance(error, RestInputValidationError):
                    response = jsonify({'message': str(error)})
                    response.status_code = 404

                elif isinstance(error, RestFeatureNotAvailable):
                    response = jsonify({'message': str(error)})
                    response.status_code = 501

                elif isinstance(error, HTTPException) and error.code == 404:
                    from aiida.restapi.common.utils import list_routes

                    response = jsonify({
                        'message': 'The requested URL is not found on the server.',
                        'available_endpoints': list_routes()
                    })
                    response.status_code = 404

                # Generic server-side error (not to make the api crash if an
                # unhandled exception is raised. Caution is never enough!!)
                else:
                    response = jsonify({'message': str(error)})
                    response.status_code = 500

                return response

        else:
            pass


class AiidaApi(Api):
    """
    AiiDA customized version of the flask_restful Api class
    """

    def __init__(self, app=None, **kwargs):
        """
        The need to have a special constructor is to include directly the
        addition of resources with the parameters required to initialize the
        resource classes.

        :param kwargs: parameters to be passed to the resources for
          configuration and PREFIX
        """

        from aiida.restapi.resources import ProcessNode, CalcJobNode, Computer, User, Group, Node, ServerInfo

        self.app = app

        super().__init__(app=app, prefix=kwargs['PREFIX'], catch_all_404s=True)

        self.add_resource(
            ServerInfo,
            '/',
            '/server/',
            '/server/endpoints/',
            endpoint='server',
            strict_slashes=False,
            resource_class_kwargs=kwargs
        )

        ## Add resources and endpoints to the api
        self.add_resource(
            Computer,
            # supported urls
            '/computers/',
            '/computers/page/',
            '/computers/page/<int:page>/',
            '/computers/<id>/',
            '/computers/projectable_properties/',
            endpoint='computers',
            strict_slashes=False,
            resource_class_kwargs=kwargs
        )

        self.add_resource(
            Node,
            '/nodes/',
            '/nodes/projectable_properties/',
            '/nodes/statistics/',
            '/nodes/full_types/',
            '/nodes/full_types_count/',
            '/nodes/download_formats/',
            '/nodes/page/',
            '/nodes/page/<int:page>/',
            '/nodes/<id>/',
            '/nodes/<id>/links/incoming/',
            '/nodes/<id>/links/incoming/page/',
            '/nodes/<id>/links/incoming/page/<int:page>/',
            '/nodes/<id>/links/outgoing/',
            '/nodes/<id>/links/outgoing/page/',
            '/nodes/<id>/links/outgoing/page/<int:page>/',
            '/nodes/<id>/links/tree/',
            '/nodes/<id>/contents/attributes/',
            '/nodes/<id>/contents/extras/',
            '/nodes/<id>/contents/derived_properties/',
            '/nodes/<id>/contents/comments/',
            '/nodes/<id>/repo/list/',
            '/nodes/<id>/repo/contents/',
            '/nodes/<id>/download/',
            endpoint='nodes',
            strict_slashes=False,
            resource_class_kwargs=kwargs
        )

        self.add_resource(
            ProcessNode,
            '/processes/projectable_properties/',
            '/processes/<id>/report/',
            endpoint='processes',
            strict_slashes=False,
            resource_class_kwargs=kwargs
        )

        self.add_resource(
            CalcJobNode,
            '/calcjobs/<id>/input_files/',
            '/calcjobs/<id>/output_files/',
            endpoint='calcjobs',
            strict_slashes=False,
            resource_class_kwargs=kwargs
        )

        self.add_resource(
            User,
            '/users/',
            '/users/projectable_properties/',
            '/users/page/',
            '/users/page/<int:page>/',
            '/users/<id>/',
            endpoint='users',
            strict_slashes=False,
            resource_class_kwargs=kwargs
        )

        self.add_resource(
            Group,
            '/groups/',
            '/groups/projectable_properties/',
            '/groups/page/',
            '/groups/page/<int:page>/',
            '/groups/<id>/',
            endpoint='groups',
            strict_slashes=False,
            resource_class_kwargs=kwargs
        )

    def handle_error(self, e):
        """
        this method handles the 404 "URL not found" exception and return custom message
        :param e: raised exception
        :return: list of available endpoints
        """
        raise e



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Migration to add the `extras` JSONB column to the `DbGroup` model."""
# pylint: disable=invalid-name
import django.contrib.postgres.fields.jsonb
from django.db import migrations
from aiida.backends.djsite.db.migrations import upgrade_schema_version

REVISION = '1.0.45'
DOWN_REVISION = '1.0.44'


class Migration(migrations.Migration):
    """Migrate to add the extras column to the dbgroup table."""
    dependencies = [
        ('db', '0044_dbgroup_type_string'),
    ]

    operations = [
        migrations.AddField(
            model_name='dbgroup',
            name='extras',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict, null=False),
        ),
        upgrade_schema_version(REVISION, DOWN_REVISION),
    ]



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Click parameter types for paths."""
import os
# See https://stackoverflow.com/a/41217363/1069467
import urllib.request
import urllib.error
from socket import timeout
import click

URL_TIMEOUT_SECONDS = 10


def _check_timeout_seconds(timeout_seconds):
    """Raise if timeout is not within range [0;60]"""
    try:
        timeout_seconds = int(timeout_seconds)
    except ValueError:
        raise TypeError(f'timeout_seconds should be an integer but got: {type(timeout_seconds)}')

    if timeout_seconds < 0 or timeout_seconds > 60:
        raise ValueError('timeout_seconds needs to be in the range [0;60].')

    return timeout_seconds


class AbsolutePathParamType(click.Path):
    """
    The ParamType for identifying absolute Paths (derived from click.Path).
    """

    name = 'AbsolutePath'

    def convert(self, value, param, ctx):
        value = os.path.expanduser(value)
        newval = super().convert(value, param, ctx)
        if not os.path.isabs(newval):
            raise click.BadParameter('path must be absolute')
        return newval

    def __repr__(self):
        return 'ABSOLUTEPATH'


class AbsolutePathOrEmptyParamType(AbsolutePathParamType):
    """
    The ParamType for identifying absolute Paths, accepting also empty paths.
    """

    name = 'AbsolutePathEmpty'

    def convert(self, value, param, ctx):
        if not value:
            return value
        return super().convert(value, param, ctx)

    def __repr__(self):
        return 'ABSOLUTEPATHEMPTY'


class PathOrUrl(click.Path):
    """Extension of click's Path-type to include URLs.

    A PathOrUrl can either be a `click.Path`-type or a URL.

    :param int timeout_seconds: Maximum timeout accepted for URL response.
        Must be an integer in the range [0;60].
    """

    # pylint: disable=protected-access

    name = 'PathOrUrl'

    def __init__(self, timeout_seconds=URL_TIMEOUT_SECONDS, **kwargs):
        super().__init__(**kwargs)

        self.timeout_seconds = _check_timeout_seconds(timeout_seconds)

    def convert(self, value, param, ctx):
        """Overwrite `convert`
        Check first if `click.Path`-type, then check if URL.
        """
        try:
            # Check if `click.Path`-type
            return super().convert(value, param, ctx)
        except click.exceptions.BadParameter:
            # Check if URL
            return self.checks_url(value, param, ctx)

    def checks_url(self, url, param, ctx):
        """Check whether URL is reachable within timeout."""
        try:
            urllib.request.urlopen(url, timeout=self.timeout_seconds)
        except (urllib.error.URLError, urllib.error.HTTPError, timeout):
            self.fail(
                '{0} "{1}" could not be reached within {2} s.\n'
                'Is it a valid {3} or URL?'.format(
                    self.path_type, click._compat.filename_to_ui(url), self.timeout_seconds, self.name
                ), param, ctx
            )

        return url


class FileOrUrl(click.File):
    """Extension of click's File-type to include URLs.

    Returns handle either to local file or to remote file fetched from URL.

    :param int timeout_seconds: Maximum timeout accepted for URL response.
        Must be an integer in the range [0;60].
    """

    name = 'FileOrUrl'

    # pylint: disable=protected-access

    def __init__(self, timeout_seconds=URL_TIMEOUT_SECONDS, **kwargs):
        super().__init__(**kwargs)

        self.timeout_seconds = _check_timeout_seconds(timeout_seconds)

    def convert(self, value, param, ctx):
        """Return file handle.
        """
        try:
            # Check if `click.File`-type
            return super().convert(value, param, ctx)
        except click.exceptions.BadParameter:
            # Check if URL
            handle = self.get_url(value, param, ctx)
        return handle

    def get_url(self, url, param, ctx):
        """Retrieve file from URL."""
        try:
            return urllib.request.urlopen(url, timeout=self.timeout_seconds)
        except (urllib.error.URLError, urllib.error.HTTPError, timeout):
            self.fail(
                '"{0}" could not be reached within {1} s.\n'
                'Is it a valid {2} or URL?.'.format(click._compat.filename_to_ui(url), self.timeout_seconds, self.name),
                param, ctx
            )



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=import-error,no-name-in-module
"""Module to manage logs for the SQLA backend."""

from sqlalchemy import ForeignKey
from sqlalchemy.dialects.postgresql import UUID, JSONB
from sqlalchemy.orm import relationship, backref
from sqlalchemy.schema import Column
from sqlalchemy.types import Integer, DateTime, String, Text

from aiida.backends.sqlalchemy.models.base import Base
from aiida.common import timezone
from aiida.common.utils import get_new_uuid


class DbLog(Base):
    """Class to store logs using SQLA backend."""
    __tablename__ = 'db_dblog'

    id = Column(Integer, primary_key=True)  # pylint: disable=invalid-name
    uuid = Column(UUID(as_uuid=True), default=get_new_uuid, unique=True)
    time = Column(DateTime(timezone=True), default=timezone.now)
    loggername = Column(String(255), index=True)
    levelname = Column(String(255), index=True)
    dbnode_id = Column(
        Integer, ForeignKey('db_dbnode.id', deferrable=True, initially='DEFERRED', ondelete='CASCADE'), nullable=False
    )
    message = Column(Text(), nullable=True)
    _metadata = Column('metadata', JSONB)

    dbnode = relationship('DbNode', backref=backref('dblogs', passive_deletes='all', cascade='merge'))

    def __init__(self, time, loggername, levelname, dbnode_id, **kwargs):
        """Setup initial value for the class attributes."""
        if 'uuid' in kwargs:
            self.uuid = kwargs['uuid']
        if 'message' in kwargs:
            self.message = kwargs['message']
        if 'metadata' in kwargs:
            self._metadata = kwargs['metadata'] or {}
        else:
            self._metadata = {}

        self.time = time
        self.loggername = loggername
        self.levelname = levelname
        self.dbnode_id = dbnode_id

    def __str__(self):
        return f'DbLog: {self.levelname} for node {self.dbnode.id}: {self.message}'



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Utilities for the implementation of the SqlAlchemy backend."""

import contextlib

# pylint: disable=import-error,no-name-in-module
from sqlalchemy import inspect
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.attributes import flag_modified

from aiida.backends.sqlalchemy import get_scoped_session
from aiida.common import exceptions

IMMUTABLE_MODEL_FIELDS = {'id', 'pk', 'uuid', 'node_type'}


class ModelWrapper:
    """Wrap a database model instance to correctly update and flush the data model when getting or setting a field.

    If the model is not stored, the behavior of the get and set attributes is unaltered. However, if the model is
    stored, which is to say, it has a primary key, the `getattr` and `setattr` are modified as follows:

    * `getattr`: if the item corresponds to a mutable model field, the model instance is refreshed first
    * `setattr`: if the item corresponds to a mutable model field, changes are flushed after performing the change
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, model, auto_flush=()):
        """Construct the ModelWrapper.

        :param model: the database model instance to wrap
        :param auto_flush: an optional tuple of database model fields that are always to be flushed, in addition to
            the field that corresponds to the attribute being set through `__setattr__`.
        """
        super().__init__()
        # Have to do it this way because we overwrite __setattr__
        object.__setattr__(self, '_model', model)
        object.__setattr__(self, '_auto_flush', auto_flush)

    def __getattr__(self, item):
        """Get an attribute of the model instance.

        If the model is saved in the database, the item corresponds to a mutable model field and the current scope is
        not in an open database connection, then the field's value is first refreshed from the database.

        :param item: the name of the model field
        :return: the value of the model's attribute
        """
        # Python 3's implementation of copy.copy does not call __init__ on the new object
        # but manually restores attributes instead. Make sure we never get into a recursive
        # loop by protecting the only special variable here: _model
        if item == '_model':
            raise AttributeError()

        if self.is_saved() and self._is_mutable_model_field(item) and not self._in_transaction():
            self._ensure_model_uptodate(fields=(item,))

        return getattr(self._model, item)

    def __setattr__(self, key, value):
        """Set the attribute on the model instance.

        If the field being set is a mutable model field and the model is saved, the changes are flushed.

        :param key: the name of the model field
        :param value: the value to set
        """
        setattr(self._model, key, value)
        if self.is_saved() and self._is_mutable_model_field(key):
            fields = set((key,) + self._auto_flush)
            self._flush(fields=fields)

    def is_saved(self):
        """Retun whether the wrapped model instance is saved in the database.

        :return: boolean, True if the model is saved in the database, False otherwise
        """
        return self._model.id is not None

    def save(self):
        """Store the model instance.

        .. note:: If one is currently in a transaction, this method is a no-op.

        :raises `aiida.common.IntegrityError`: if a database integrity error is raised during the save.
        """
        try:
            commit = not self._in_transaction()
            self._model.save(commit=commit)
        except IntegrityError as exception:
            self._model.session.rollback()
            raise exceptions.IntegrityError(str(exception))

    def _is_mutable_model_field(self, field):
        """Return whether the field is a mutable field of the model.

        :return: boolean, True if the field is a model field and is not in the `IMMUTABLE_MODEL_FIELDS` set.
        """
        if field in IMMUTABLE_MODEL_FIELDS:
            return False

        return self._is_model_field(field)

    def _is_model_field(self, field):
        """Return whether the field is a field of the model.

        :return: boolean, True if the field is a model field, False otherwise.
        """
        return inspect(self._model.__class__).has_property(field)

    def _flush(self, fields=()):
        """Flush the fields of the model to the database.

        .. note:: If the wrapped model is not actually saved in the database yet, this method is a no-op.

        :param fields: the model fields whose current value to flush to the database
        """
        if self.is_saved():
            for field in fields:
                flag_modified(self._model, field)

            self.save()

    def _ensure_model_uptodate(self, fields=None):
        """Refresh all fields of the wrapped model instance by fetching the current state of the database instance.

        :param fields: optionally refresh only these fields, if `None` all fields are refreshed.
        """
        self._model.session.expire(self._model, attribute_names=fields)

    @staticmethod
    def _in_transaction():
        """Return whether the current scope is within an open database transaction.

        :return: boolean, True if currently in open transaction, False otherwise.
        """
        return get_scoped_session().transaction.nested


@contextlib.contextmanager
def disable_expire_on_commit(session):
    """Context manager that disables expire_on_commit and restores the original value on exit

    :param session: The SQLA session
    :type session: :class:`sqlalchemy.orm.session.Session`
    """
    current_value = session.expire_on_commit
    session.expire_on_commit = False
    try:
        yield session
    finally:
        session.expire_on_commit = current_value



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name
"""Database migration."""
from django.db import migrations
from aiida.backends.djsite.db.migrations import upgrade_schema_version

REVISION = '1.0.9'
DOWN_REVISION = '1.0.8'


class Migration(migrations.Migration):
    """Database migration."""

    dependencies = [
        ('db', '0008_code_hidden_to_extra'),
    ]

    operations = [
        # The base Data types Bool, Float, Int and Str have been moved in the source code, which means that their
        # module path changes, which determines the plugin type string which is stored in the databse.
        # The type string now will have a type string prefix that is unique to each sub type.
        migrations.RunSQL(
            """
            UPDATE db_dbnode SET type = 'data.bool.Bool.' WHERE type = 'data.base.Bool.';
            UPDATE db_dbnode SET type = 'data.float.Float.' WHERE type = 'data.base.Float.';
            UPDATE db_dbnode SET type = 'data.int.Int.' WHERE type = 'data.base.Int.';
            UPDATE db_dbnode SET type = 'data.str.Str.' WHERE type = 'data.base.Str.';
            UPDATE db_dbnode SET type = 'data.list.List.' WHERE type = 'data.base.List.';
        """
        ),
        upgrade_schema_version(REVISION, DOWN_REVISION)
    ]



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Convenience functions to add awaitables to the Context of a WorkChain."""
from .awaitable import construct_awaitable, AwaitableAction

__all__ = ('ToContext', 'assign_', 'append_')

ToContext = dict


def assign_(target):
    """
    Convenience function that will construct an Awaitable for a given class instance
    with the context action set to ASSIGN. When the awaitable target is completed
    it will be assigned to the context for a key that is to be defined later

    :param target: an instance of a Process or Awaitable

    :returns: the awaitable
    :rtype: Awaitable
    """
    awaitable = construct_awaitable(target)
    awaitable.action = AwaitableAction.ASSIGN
    return awaitable


def append_(target):
    """
    Convenience function that will construct an Awaitable for a given class instance
    with the context action set to APPEND. When the awaitable target is completed
    it will be appended to a list in the context for a key that is to be defined later

    :param target: an instance of a Process or Awaitable

    :returns: the awaitable
    :rtype: Awaitable
    """
    awaitable = construct_awaitable(target)
    awaitable.action = AwaitableAction.APPEND
    return awaitable



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
A module defining hydrogen-like orbitals that are real valued (rather than
complex-valued).
"""

from aiida.common.exceptions import ValidationError, InputValidationError

from .orbital import Orbital, validate_len3_list_or_none, validate_float_or_none


def validate_l(value):
    """
    Validate the value of the angular momentum
    """
    if not isinstance(value, int):
        raise ValidationError('angular momentum (l) must be integer')

    if value < -5 or value > 3:
        raise ValidationError('angular momentum (l) must be between -5 and 3')

    return value


def validate_m(value):
    """
    Validate the value of the magnetic number
    """
    if not isinstance(value, int):
        raise ValidationError('magnetic number (m) must be integer')

    # without knowing l we cannot validate the value of m. We will call an additional function
    # in the validate function

    return value


def validate_kind_name(value):
    """
    Validate the value of the kind_name
    """
    if value is not None and not isinstance(value, str):
        raise ValidationError('kind_name must be a string')

    return value


def validate_n(value):
    """
    Validate the value of the number of radial nodes
    """
    if not isinstance(value, int):
        raise ValidationError('number of radial nodes (n) must be integer')

    if value < 0 or value > 2:
        raise ValidationError('number of radial nodes (n) must be between 0 and 2')

    return value


def validate_spin(value):
    """
    Validate the value of the spin
    """
    if not isinstance(value, int):
        raise ValidationError('spin must be integer')

    if value < -1 or value > 1:
        raise ValidationError('spin must be among -1, 1 or 0 (undefined)')

    return value


CONVERSION_DICT = {
    'S': {
        'S': {
            'angular_momentum': 0,
            'magnetic_number': 0
        }
    },
    'P': {
        'PZ': {
            'angular_momentum': 1,
            'magnetic_number': 0
        },
        'PX': {
            'angular_momentum': 1,
            'magnetic_number': 1
        },
        'PY': {
            'angular_momentum': 1,
            'magnetic_number': 2
        },
    },
    'D': {
        'DZ2': {
            'angular_momentum': 2,
            'magnetic_number': 0
        },
        'DXZ': {
            'angular_momentum': 2,
            'magnetic_number': 1
        },
        'DYZ': {
            'angular_momentum': 2,
            'magnetic_number': 2
        },
        'DX2-Y2': {
            'angular_momentum': 2,
            'magnetic_number': 3
        },
        'DXY': {
            'angular_momentum': 2,
            'magnetic_number': 4
        },
    },
    'F': {
        'FZ3': {
            'angular_momentum': 3,
            'magnetic_number': 0
        },
        'FZX2': {
            'angular_momentum': 3,
            'magnetic_number': 1
        },
        'FYZ2': {
            'angular_momentum': 3,
            'magnetic_number': 2
        },
        'FZ(X2-Y2)': {
            'angular_momentum': 3,
            'magnetic_number': 3
        },
        'FXYZ': {
            'angular_momentum': 3,
            'magnetic_number': 4
        },
        'FX(X2-3Y2)': {
            'angular_momentum': 3,
            'magnetic_number': 5
        },
        'FY(3X2-Y2)': {
            'angular_momentum': 3,
            'magnetic_number': 6
        },
    },
    'SP': {
        'SP-1': {
            'angular_momentum': -1,
            'magnetic_number': 0
        },
        'SP-2': {
            'angular_momentum': -1,
            'magnetic_number': 1
        },
    },
    'SP2': {
        'SP2-1': {
            'angular_momentum': -2,
            'magnetic_number': 0
        },
        'SP2-2': {
            'angular_momentum': -2,
            'magnetic_number': 1
        },
        'SP2-3': {
            'angular_momentum': -2,
            'magnetic_number': 2
        },
    },
    'SP3': {
        'SP3-1': {
            'angular_momentum': -3,
            'magnetic_number': 0
        },
        'SP3-2': {
            'angular_momentum': -3,
            'magnetic_number': 1
        },
        'SP3-3': {
            'angular_momentum': -3,
            'magnetic_number': 2
        },
        'SP3-4': {
            'angular_momentum': -3,
            'magnetic_number': 3
        },
    },
    'SP3D': {
        'SP3D-1': {
            'angular_momentum': -4,
            'magnetic_number': 0
        },
        'SP3D-2': {
            'angular_momentum': -4,
            'magnetic_number': 1
        },
        'SP3D-3': {
            'angular_momentum': -4,
            'magnetic_number': 2
        },
        'SP3D-4': {
            'angular_momentum': -4,
            'magnetic_number': 3
        },
        'SP3D-5': {
            'angular_momentum': -4,
            'magnetic_number': 4
        },
    },
    'SP3D2': {
        'SP3D2-1': {
            'angular_momentum': -5,
            'magnetic_number': 0
        },
        'SP3D2-2': {
            'angular_momentum': -5,
            'magnetic_number': 1
        },
        'SP3D2-3': {
            'angular_momentum': -5,
            'magnetic_number': 2
        },
        'SP3D2-4': {
            'angular_momentum': -5,
            'magnetic_number': 3
        },
        'SP3D2-5': {
            'angular_momentum': -5,
            'magnetic_number': 4
        },
        'SP3D2-6': {
            'angular_momentum': -5,
            'magnetic_number': 5
        },
    }
}


class RealhydrogenOrbital(Orbital):
    """
    Orbitals for hydrogen, largely follows the conventions used by wannier90
    Object to handle the generation of real hydrogen orbitals and their
    hybrids, has methods for producing s, p, d, f, and sp, sp2, sp3, sp3d,
    sp3d2 hybrids. This method does not deal with the cannonical hydrogen
    orbitals which contain imaginary components.

    The orbitals described here are chiefly concerned with the symmetric
    aspects of the oribitals without the context of space. Therefore
    diffusitivity, position and atomic labels should be handled in the
    OrbitalData class.

    Following the notation of table 3.1, 3.2 of Wannier90 user guide
    http://www.wannier.org/doc/user_guide.pdf
    A brief description of what is meant by each of these labels:

    :param radial_nodes: the number of radial nodes (or inflections) if no
                         radial nodes are supplied, defaults to 0
    :param angular_momentum: Angular quantum number, using real orbitals
    :param magnetic_number: Magnetic quantum number, using real orbitals
    :param spin: spin, up (1) down (-1) or unspecified (0)

    The conventions regarding L and M correpsond to those used in
    wannier90 for all L greater than 0 the orbital is not hyrbridized
    see table 3.1 and for L less than 0 the orbital is hybridized see
    table 3.2. M then indexes all the possible orbitals from 0 to 2L for
    L > 0 and from 0 to (-L) for L < 0.
    """
    _base_fields_required = tuple(
        list(Orbital._base_fields_required) +
        [('angular_momentum', validate_l), ('magnetic_number', validate_m), ('radial_nodes', validate_n)]
    )

    _base_fields_optional = tuple(
        list(Orbital._base_fields_optional) + [
            ('kind_name', validate_kind_name, None),
            ('spin', validate_spin, 0),
            ('x_orientation', validate_len3_list_or_none, None),
            ('z_orientation', validate_len3_list_or_none, None),
            ('spin_orientation', validate_len3_list_or_none, None),
            ('diffusivity', validate_float_or_none, None),
        ]
    )

    def __str__(self):
        orb_dict = self.get_orbital_dict()
        try:
            orb_name = self.get_name_from_quantum_numbers(
                orb_dict['angular_momentum'], magnetic_number=orb_dict['magnetic_number']
            )
            position_string = '{:.4f},{:.4f},{:.4f}'.format(
                orb_dict['position'][0], orb_dict['position'][1], orb_dict['position'][2]
            )
            out_string = 'r{} {} orbital {} @ {}'.format(
                orb_dict['radial_nodes'], orb_name,
                "for kind '{}'".format(orb_dict['kind_name']) if orb_dict['kind_name'] else '', position_string
            )
        except KeyError:
            # Should not happen, but we want it not to crash in __str__
            out_string = '(not all parameters properly set)'
        return out_string

    def _validate_keys(self, input_dict):
        """
        Validates the keys otherwise raise ValidationError. Does basic
        validation from the parent followed by validations for the
        quantum numbers. Raises exceptions should the input_dict fail the
        valiation or if it contains any unsupported keywords.

        :param input_dict: the dictionary of keys to be validated
        :return validated_dict: a validated dictionary
        """
        validated_dict = super()._validate_keys(input_dict)

        # Validate m knowing the value of l
        angular_momentum = validated_dict['angular_momentum']  # l quantum number, must be there
        magnetic_number = validated_dict['magnetic_number']  # m quantum number, must be there
        if angular_momentum >= 0:
            accepted_range = [0, 2 * angular_momentum]
        else:
            accepted_range = [0, -angular_momentum]
        if magnetic_number < min(accepted_range) or magnetic_number > max(accepted_range):
            raise ValidationError(
                f'the magnetic number must be in the range [{min(accepted_range)}, {max(accepted_range)}]'
            )

        # Check if it is a known combination
        try:
            self.get_name_from_quantum_numbers(
                validated_dict['angular_momentum'], magnetic_number=validated_dict['magnetic_number']
            )
        except ValidationError:
            raise ValidationError('Invalid angular momentum magnetic number combination')

        return validated_dict

    @classmethod
    def get_name_from_quantum_numbers(cls, angular_momentum, magnetic_number=None):
        """
        Returns the orbital_name correponding to the angular_momentum alone,
        or to both angular_number with magnetic_number. For example using
        angular_momentum=1 and magnetic_number=1 will return "Px"
        """
        orbital_name = [
            x for x in CONVERSION_DICT
            if any([CONVERSION_DICT[x][y]['angular_momentum'] == angular_momentum for y in CONVERSION_DICT[x]])
        ]
        if not orbital_name:
            raise InputValidationError(
                f'No orbital name corresponding to the angular_momentum {angular_momentum} could be found'
            )
        if magnetic_number is not None:
            # finds angular momentum
            orbital_name = orbital_name[0]
            orbital_name = [
                x for x in CONVERSION_DICT[orbital_name]
                if CONVERSION_DICT[orbital_name][x]['magnetic_number'] == magnetic_number
            ]

            if not orbital_name:
                raise InputValidationError(
                    f'No orbital name corresponding to the magnetic_number {magnetic_number} could be found'
                )
        return orbital_name[0]

    @classmethod
    def get_quantum_numbers_from_name(cls, name):
        """
        Returns all the angular and magnetic numbers corresponding to name.
        For example, using "P" as name will return all quantum numbers
        associated with a "P" orbital, while "Px" will return only one set
        of quantum numbers, the ones associated with "Px"
        """
        name = name.upper()
        list_of_dicts = [CONVERSION_DICT[x][y] for x in CONVERSION_DICT for y in CONVERSION_DICT[x] if name in (y, x)]
        if not list_of_dicts:
            raise InputValidationError('Invalid choice of projection name')
        return list_of_dicts


# # Can move these later, but for now I want to have these functions handy
# def cart2sph(x, y, z):
#     """
#     Converts cartesian to unit vector spherical coordinates
#
#     :param x, y, z: absolute cartesian coordinates
#     :return: [r, theta, phi] unit vector in spherical coordinates
#     """
#     import numpy as np
#     coord_cart = np.array([x, y, z])
#     unit_cart = coord_cart/np.linalg.norm(coord_cart)
#     x,y,z = unit_cart # for clarity
#     r = np.linalg.norm(unit_cart)  # should always be 1.0
#     theta = np.arctan2(y,x)
#     phi = np.arctan2(z,r)
#     # note that this method will not return the r value, which is not stored
#     return [theta, phi]

# def sph2cart(theta, phi):
#     """
#     Converts spherical coordinates r, theta, phi to unit vector in cartesian
#     coordinates. r is not accepted as input because only unit vectors are
#     considered here.
#
#     : theta, phi: spherical coordinates vector
#     :return: x, y, z unit vector  in cartesian absolute coordinates
#     """
#     import numpy as np
#     # code is done in a few more steps than necessary, to make logic explicit
#     r = 1
#     x = r*np.cos(theta)*np.sin(phi)
#     y = r*np.sin(theta)*np.sin(phi)
#     z = r*np.cos(theta)
#     x = x/r
#     y = y/r
#     z = z/r
#     return [x, y, z]
